# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_rpc/benchmark.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_rpc/benchmark.proto',
  package='pw.rpc',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x16pw_rpc/benchmark.proto\x12\x06pw.rpc\"\x1a\n\x07Payload\x12\x0f\n\x07payload\x18\x01 \x01(\x0c\x32u\n\tBenchmark\x12-\n\tUnaryEcho\x12\x0f.pw.rpc.Payload\x1a\x0f.pw.rpc.Payload\x12\x39\n\x11\x42idirectionalEcho\x12\x0f.pw.rpc.Payload\x1a\x0f.pw.rpc.Payload(\x01\x30\x01\x62\x06proto3'
)




_PAYLOAD = _descriptor.Descriptor(
  name='Payload',
  full_name='pw.rpc.Payload',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='payload', full_name='pw.rpc.Payload.payload', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=34,
  serialized_end=60,
)

DESCRIPTOR.message_types_by_name['Payload'] = _PAYLOAD
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Payload = _reflection.GeneratedProtocolMessageType('Payload', (_message.Message,), {
  'DESCRIPTOR' : _PAYLOAD,
  '__module__' : 'pw_rpc.benchmark_pb2'
  # @@protoc_insertion_point(class_scope:pw.rpc.Payload)
  })
_sym_db.RegisterMessage(Payload)



_BENCHMARK = _descriptor.ServiceDescriptor(
  name='Benchmark',
  full_name='pw.rpc.Benchmark',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=62,
  serialized_end=179,
  methods=[
  _descriptor.MethodDescriptor(
    name='UnaryEcho',
    full_name='pw.rpc.Benchmark.UnaryEcho',
    index=0,
    containing_service=None,
    input_type=_PAYLOAD,
    output_type=_PAYLOAD,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BidirectionalEcho',
    full_name='pw.rpc.Benchmark.BidirectionalEcho',
    index=1,
    containing_service=None,
    input_type=_PAYLOAD,
    output_type=_PAYLOAD,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_BENCHMARK)

DESCRIPTOR.services_by_name['Benchmark'] = _BENCHMARK

# @@protoc_insertion_point(module_scope)
