/*
 * Copyright 2022 The Pigweed Authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 *     https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
body {
  background-color: rgb(46, 46, 46);
  color: #ffffff;
  overflow: hidden;
  margin: 0;
}

.table-container {
  display: table;
  width: 100%;
  border-spacing: 30px 0px;
}

.log-header {
  font-size: 18px;
  font-family: monospace;
}

.log-container {
  width: 100%;
  height: calc(100vh - 50px);
  overflow-y: auto;
  border-top: 1px solid #DDD;
  font-size: 18px;
  font-family: monospace;
}

.log-header {
  width: 100%;
  font-weight: bold;
  display: table-row;
}

.log-container .row>span {
  display: table-cell;
  padding: 20px 18px;

}

.log-header>span {
  text-transform: capitalize;
  overflow: hidden;
    display: inline-block;
    margin-left: 30px;
}

.log-entry {
  display: table-row;
}

.log-entry>span {
  display: table-cell;
  overflow: hidden;
  text-overflow: ellipsis;
}

.log-entry .msg {
  flex: 1;
}
