# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_metric_proto/metric_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_metric_proto/metric_service.proto',
  package='pw.metric.proto',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n$pw_metric_proto/metric_service.proto\x12\x0fpw.metric.proto\"`\n\x06Metric\x12\x12\n\ntoken_path\x18\x01 \x03(\x07\x12\x13\n\x0bstring_path\x18\x02 \x03(\t\x12\x12\n\x08\x61s_float\x18\x03 \x01(\x02H\x00\x12\x10\n\x06\x61s_int\x18\x04 \x01(\rH\x00\x42\x07\n\x05value\"9\n\rMetricRequest\x12(\n\x07metrics\x18\x01 \x03(\x0b\x32\x17.pw.metric.proto.Metric\":\n\x0eMetricResponse\x12(\n\x07metrics\x18\x01 \x03(\x0b\x32\x17.pw.metric.proto.Metric2[\n\rMetricService\x12J\n\x03Get\x12\x1e.pw.metric.proto.MetricRequest\x1a\x1f.pw.metric.proto.MetricResponse\"\x00\x30\x01\x62\x06proto3'
)




_METRIC = _descriptor.Descriptor(
  name='Metric',
  full_name='pw.metric.proto.Metric',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='token_path', full_name='pw.metric.proto.Metric.token_path', index=0,
      number=1, type=7, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='string_path', full_name='pw.metric.proto.Metric.string_path', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='as_float', full_name='pw.metric.proto.Metric.as_float', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='as_int', full_name='pw.metric.proto.Metric.as_int', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='value', full_name='pw.metric.proto.Metric.value',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=57,
  serialized_end=153,
)


_METRICREQUEST = _descriptor.Descriptor(
  name='MetricRequest',
  full_name='pw.metric.proto.MetricRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metrics', full_name='pw.metric.proto.MetricRequest.metrics', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=155,
  serialized_end=212,
)


_METRICRESPONSE = _descriptor.Descriptor(
  name='MetricResponse',
  full_name='pw.metric.proto.MetricResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metrics', full_name='pw.metric.proto.MetricResponse.metrics', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=214,
  serialized_end=272,
)

_METRIC.oneofs_by_name['value'].fields.append(
  _METRIC.fields_by_name['as_float'])
_METRIC.fields_by_name['as_float'].containing_oneof = _METRIC.oneofs_by_name['value']
_METRIC.oneofs_by_name['value'].fields.append(
  _METRIC.fields_by_name['as_int'])
_METRIC.fields_by_name['as_int'].containing_oneof = _METRIC.oneofs_by_name['value']
_METRICREQUEST.fields_by_name['metrics'].message_type = _METRIC
_METRICRESPONSE.fields_by_name['metrics'].message_type = _METRIC
DESCRIPTOR.message_types_by_name['Metric'] = _METRIC
DESCRIPTOR.message_types_by_name['MetricRequest'] = _METRICREQUEST
DESCRIPTOR.message_types_by_name['MetricResponse'] = _METRICRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Metric = _reflection.GeneratedProtocolMessageType('Metric', (_message.Message,), {
  'DESCRIPTOR' : _METRIC,
  '__module__' : 'pw_metric_proto.metric_service_pb2'
  # @@protoc_insertion_point(class_scope:pw.metric.proto.Metric)
  })
_sym_db.RegisterMessage(Metric)

MetricRequest = _reflection.GeneratedProtocolMessageType('MetricRequest', (_message.Message,), {
  'DESCRIPTOR' : _METRICREQUEST,
  '__module__' : 'pw_metric_proto.metric_service_pb2'
  # @@protoc_insertion_point(class_scope:pw.metric.proto.MetricRequest)
  })
_sym_db.RegisterMessage(MetricRequest)

MetricResponse = _reflection.GeneratedProtocolMessageType('MetricResponse', (_message.Message,), {
  'DESCRIPTOR' : _METRICRESPONSE,
  '__module__' : 'pw_metric_proto.metric_service_pb2'
  # @@protoc_insertion_point(class_scope:pw.metric.proto.MetricResponse)
  })
_sym_db.RegisterMessage(MetricResponse)



_METRICSERVICE = _descriptor.ServiceDescriptor(
  name='MetricService',
  full_name='pw.metric.proto.MetricService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=274,
  serialized_end=365,
  methods=[
  _descriptor.MethodDescriptor(
    name='Get',
    full_name='pw.metric.proto.MetricService.Get',
    index=0,
    containing_service=None,
    input_type=_METRICREQUEST,
    output_type=_METRICRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_METRICSERVICE)

DESCRIPTOR.services_by_name['MetricService'] = _METRICSERVICE

# @@protoc_insertion_point(module_scope)
