# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_software_update/tuf.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_software_update/tuf.proto',
  package='pw.software_update',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1cpw_software_update/tuf.proto\x12\x12pw.software_update\x1a\x1fgoogle/protobuf/timestamp.proto\"i\n\x12SignedRootMetadata\x12 \n\x18serialized_root_metadata\x18\x01 \x01(\x0c\x12\x31\n\nsignatures\x18\x02 \x03(\x0b\x32\x1d.pw.software_update.Signature\"s\n\x17SignedTimestampMetadata\x12%\n\x1dserialized_timestamp_metadata\x18\x01 \x01(\x0c\x12\x31\n\nsignatures\x18\x02 \x03(\x0b\x32\x1d.pw.software_update.Signature\"q\n\x16SignedSnapshotMetadata\x12$\n\x1cserialized_snapshot_metadata\x18\x01 \x01(\x0c\x12\x31\n\nsignatures\x18\x02 \x03(\x0b\x32\x1d.pw.software_update.Signature\"o\n\x15SignedTargetsMetadata\x12#\n\x1bserialized_targets_metadata\x18\x01 \x01(\x0c\x12\x31\n\nsignatures\x18\x02 \x03(\x0b\x32\x1d.pw.software_update.Signature\"\x91\x01\n\x0e\x43ommonMetadata\x12\x14\n\x0cspec_version\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\r\x12\x30\n\x07\x65xpires\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x00\x88\x01\x01\x12\x11\n\x04role\x18\x04 \x01(\tH\x01\x88\x01\x01\x42\n\n\x08_expiresB\x07\n\x05_role\"\xed\x03\n\x0cRootMetadata\x12;\n\x0f\x63ommon_metadata\x18\x01 \x01(\x0b\x32\".pw.software_update.CommonMetadata\x12\x1b\n\x13\x63onsistent_snapshot\x18\x02 \x01(\x08\x12,\n\x04keys\x18\x03 \x03(\x0b\x32\x1e.pw.software_update.KeyMapping\x12L\n\x1aroot_signature_requirement\x18\x04 \x01(\x0b\x32(.pw.software_update.SignatureRequirement\x12Q\n\x1ftimestamp_signature_requirement\x18\x05 \x01(\x0b\x32(.pw.software_update.SignatureRequirement\x12P\n\x1esnapshot_signature_requirement\x18\x06 \x01(\x0b\x32(.pw.software_update.SignatureRequirement\x12O\n\x1dtargets_signature_requirement\x18\x07 \x01(\x0b\x32(.pw.software_update.SignatureRequirementJ\x04\x08\x08\x10 J\x04\x08 \x10\x41J\x05\x08\x41\x10\x80\x02\"\xa0\x01\n\x11TimestampMetadata\x12;\n\x0f\x63ommon_metadata\x18\x01 \x01(\x0b\x32\".pw.software_update.CommonMetadata\x12;\n\x11snapshot_metadata\x18\x02 \x01(\x0b\x32 .pw.software_update.MetadataFileJ\x04\x08\x03\x10 J\x04\x08 \x10\x41J\x05\x08\x41\x10\x80\x02\"\x9e\x01\n\x10SnapshotMetadata\x12;\n\x0f\x63ommon_metadata\x18\x01 \x01(\x0b\x32\".pw.software_update.CommonMetadata\x12:\n\x10targets_metadata\x18\x02 \x03(\x0b\x32 .pw.software_update.MetadataFileJ\x04\x08\x03\x10 J\x04\x08 \x10\x41J\x05\x08\x41\x10\x80\x02\"\x97\x01\n\x0fTargetsMetadata\x12;\n\x0f\x63ommon_metadata\x18\x01 \x01(\x0b\x32\".pw.software_update.CommonMetadata\x12\x34\n\x0ctarget_files\x18\x02 \x03(\x0b\x32\x1e.pw.software_update.TargetFileJ\x04\x08\t\x10 J\x04\x08 \x10\x41J\x05\x08\x41\x10\x80\x02\"(\n\tSignature\x12\x0e\n\x06key_id\x18\x01 \x01(\x0c\x12\x0b\n\x03sig\x18\x02 \x01(\x0c\"B\n\nKeyMapping\x12\x0e\n\x06key_id\x18\x01 \x01(\x0c\x12$\n\x03key\x18\x02 \x01(\x0b\x32\x17.pw.software_update.Key\"s\n\x03Key\x12-\n\x08key_type\x18\x01 \x01(\x0e\x32\x1b.pw.software_update.KeyType\x12-\n\x06scheme\x18\x02 \x01(\x0e\x32\x1d.pw.software_update.KeyScheme\x12\x0e\n\x06keyval\x18\x03 \x01(\x0c\":\n\x14SignatureRequirement\x12\x0f\n\x07key_ids\x18\x01 \x03(\x0c\x12\x11\n\tthreshold\x18\x02 \x01(\r\"H\n\x04Hash\x12\x32\n\x08\x66unction\x18\x01 \x01(\x0e\x32 .pw.software_update.HashFunction\x12\x0c\n\x04hash\x18\x02 \x01(\x0c\"l\n\nTargetFile\x12\x11\n\tfile_name\x18\x01 \x01(\t\x12\x0e\n\x06length\x18\x02 \x01(\x04\x12(\n\x06hashes\x18\x03 \x03(\x0b\x32\x18.pw.software_update.HashJ\x04\x08\x04\x10\x10J\x04\x08\x10\x10 J\x05\x08 \x10\x80\x02\"\x8f\x01\n\x0cMetadataFile\x12\x16\n\tfile_name\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x0f\n\x07version\x18\x02 \x01(\r\x12\x13\n\x06length\x18\x03 \x01(\x04H\x01\x88\x01\x01\x12(\n\x06hashes\x18\x04 \x03(\x0b\x32\x18.pw.software_update.HashB\x0c\n\n_file_nameB\t\n\x07_length*5\n\x0cHashFunction\x12\x19\n\x15UNKNOWN_HASH_FUNCTION\x10\x00\x12\n\n\x06SHA256\x10\x01*N\n\x07KeyType\x12\x14\n\x10UNKNOWN_KEY_TYPE\x10\x00\x12\x07\n\x03RSA\x10\x01\x12\x0b\n\x07\x45\x44\x32\x35\x35\x31\x39\x10\x02\x12\x17\n\x13\x45\x43\x44SA_SHA2_NISTP256\x10\x03*u\n\tKeyScheme\x12\x16\n\x12UNKNOWN_KEY_SCHEME\x10\x00\x12\x1c\n\x18RSASSA_PSS_SHA256_SCHEME\x10\x01\x12\x12\n\x0e\x45\x44\x32\x35\x35\x31\x39_SCHEME\x10\x02\x12\x1e\n\x1a\x45\x43\x44SA_SHA2_NISTP256_SCHEME\x10\x03\x62\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,])

_HASHFUNCTION = _descriptor.EnumDescriptor(
  name='HashFunction',
  full_name='pw.software_update.HashFunction',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN_HASH_FUNCTION', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SHA256', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2276,
  serialized_end=2329,
)
_sym_db.RegisterEnumDescriptor(_HASHFUNCTION)

HashFunction = enum_type_wrapper.EnumTypeWrapper(_HASHFUNCTION)
_KEYTYPE = _descriptor.EnumDescriptor(
  name='KeyType',
  full_name='pw.software_update.KeyType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN_KEY_TYPE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RSA', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ED25519', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ECDSA_SHA2_NISTP256', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2331,
  serialized_end=2409,
)
_sym_db.RegisterEnumDescriptor(_KEYTYPE)

KeyType = enum_type_wrapper.EnumTypeWrapper(_KEYTYPE)
_KEYSCHEME = _descriptor.EnumDescriptor(
  name='KeyScheme',
  full_name='pw.software_update.KeyScheme',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN_KEY_SCHEME', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RSASSA_PSS_SHA256_SCHEME', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ED25519_SCHEME', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ECDSA_SHA2_NISTP256_SCHEME', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2411,
  serialized_end=2528,
)
_sym_db.RegisterEnumDescriptor(_KEYSCHEME)

KeyScheme = enum_type_wrapper.EnumTypeWrapper(_KEYSCHEME)
UNKNOWN_HASH_FUNCTION = 0
SHA256 = 1
UNKNOWN_KEY_TYPE = 0
RSA = 1
ED25519 = 2
ECDSA_SHA2_NISTP256 = 3
UNKNOWN_KEY_SCHEME = 0
RSASSA_PSS_SHA256_SCHEME = 1
ED25519_SCHEME = 2
ECDSA_SHA2_NISTP256_SCHEME = 3



_SIGNEDROOTMETADATA = _descriptor.Descriptor(
  name='SignedRootMetadata',
  full_name='pw.software_update.SignedRootMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='serialized_root_metadata', full_name='pw.software_update.SignedRootMetadata.serialized_root_metadata', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='signatures', full_name='pw.software_update.SignedRootMetadata.signatures', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=85,
  serialized_end=190,
)


_SIGNEDTIMESTAMPMETADATA = _descriptor.Descriptor(
  name='SignedTimestampMetadata',
  full_name='pw.software_update.SignedTimestampMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='serialized_timestamp_metadata', full_name='pw.software_update.SignedTimestampMetadata.serialized_timestamp_metadata', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='signatures', full_name='pw.software_update.SignedTimestampMetadata.signatures', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=192,
  serialized_end=307,
)


_SIGNEDSNAPSHOTMETADATA = _descriptor.Descriptor(
  name='SignedSnapshotMetadata',
  full_name='pw.software_update.SignedSnapshotMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='serialized_snapshot_metadata', full_name='pw.software_update.SignedSnapshotMetadata.serialized_snapshot_metadata', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='signatures', full_name='pw.software_update.SignedSnapshotMetadata.signatures', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=309,
  serialized_end=422,
)


_SIGNEDTARGETSMETADATA = _descriptor.Descriptor(
  name='SignedTargetsMetadata',
  full_name='pw.software_update.SignedTargetsMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='serialized_targets_metadata', full_name='pw.software_update.SignedTargetsMetadata.serialized_targets_metadata', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='signatures', full_name='pw.software_update.SignedTargetsMetadata.signatures', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=424,
  serialized_end=535,
)


_COMMONMETADATA = _descriptor.Descriptor(
  name='CommonMetadata',
  full_name='pw.software_update.CommonMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='spec_version', full_name='pw.software_update.CommonMetadata.spec_version', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='version', full_name='pw.software_update.CommonMetadata.version', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='expires', full_name='pw.software_update.CommonMetadata.expires', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='role', full_name='pw.software_update.CommonMetadata.role', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_expires', full_name='pw.software_update.CommonMetadata._expires',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_role', full_name='pw.software_update.CommonMetadata._role',
      index=1, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=538,
  serialized_end=683,
)


_ROOTMETADATA = _descriptor.Descriptor(
  name='RootMetadata',
  full_name='pw.software_update.RootMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='common_metadata', full_name='pw.software_update.RootMetadata.common_metadata', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='consistent_snapshot', full_name='pw.software_update.RootMetadata.consistent_snapshot', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='keys', full_name='pw.software_update.RootMetadata.keys', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='root_signature_requirement', full_name='pw.software_update.RootMetadata.root_signature_requirement', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp_signature_requirement', full_name='pw.software_update.RootMetadata.timestamp_signature_requirement', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='snapshot_signature_requirement', full_name='pw.software_update.RootMetadata.snapshot_signature_requirement', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='targets_signature_requirement', full_name='pw.software_update.RootMetadata.targets_signature_requirement', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=686,
  serialized_end=1179,
)


_TIMESTAMPMETADATA = _descriptor.Descriptor(
  name='TimestampMetadata',
  full_name='pw.software_update.TimestampMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='common_metadata', full_name='pw.software_update.TimestampMetadata.common_metadata', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='snapshot_metadata', full_name='pw.software_update.TimestampMetadata.snapshot_metadata', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1182,
  serialized_end=1342,
)


_SNAPSHOTMETADATA = _descriptor.Descriptor(
  name='SnapshotMetadata',
  full_name='pw.software_update.SnapshotMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='common_metadata', full_name='pw.software_update.SnapshotMetadata.common_metadata', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='targets_metadata', full_name='pw.software_update.SnapshotMetadata.targets_metadata', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1345,
  serialized_end=1503,
)


_TARGETSMETADATA = _descriptor.Descriptor(
  name='TargetsMetadata',
  full_name='pw.software_update.TargetsMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='common_metadata', full_name='pw.software_update.TargetsMetadata.common_metadata', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='target_files', full_name='pw.software_update.TargetsMetadata.target_files', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1506,
  serialized_end=1657,
)


_SIGNATURE = _descriptor.Descriptor(
  name='Signature',
  full_name='pw.software_update.Signature',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key_id', full_name='pw.software_update.Signature.key_id', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sig', full_name='pw.software_update.Signature.sig', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1659,
  serialized_end=1699,
)


_KEYMAPPING = _descriptor.Descriptor(
  name='KeyMapping',
  full_name='pw.software_update.KeyMapping',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key_id', full_name='pw.software_update.KeyMapping.key_id', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='key', full_name='pw.software_update.KeyMapping.key', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1701,
  serialized_end=1767,
)


_KEY = _descriptor.Descriptor(
  name='Key',
  full_name='pw.software_update.Key',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key_type', full_name='pw.software_update.Key.key_type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='scheme', full_name='pw.software_update.Key.scheme', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='keyval', full_name='pw.software_update.Key.keyval', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1769,
  serialized_end=1884,
)


_SIGNATUREREQUIREMENT = _descriptor.Descriptor(
  name='SignatureRequirement',
  full_name='pw.software_update.SignatureRequirement',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key_ids', full_name='pw.software_update.SignatureRequirement.key_ids', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='threshold', full_name='pw.software_update.SignatureRequirement.threshold', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1886,
  serialized_end=1944,
)


_HASH = _descriptor.Descriptor(
  name='Hash',
  full_name='pw.software_update.Hash',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='function', full_name='pw.software_update.Hash.function', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='hash', full_name='pw.software_update.Hash.hash', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1946,
  serialized_end=2018,
)


_TARGETFILE = _descriptor.Descriptor(
  name='TargetFile',
  full_name='pw.software_update.TargetFile',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='file_name', full_name='pw.software_update.TargetFile.file_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='length', full_name='pw.software_update.TargetFile.length', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='hashes', full_name='pw.software_update.TargetFile.hashes', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2020,
  serialized_end=2128,
)


_METADATAFILE = _descriptor.Descriptor(
  name='MetadataFile',
  full_name='pw.software_update.MetadataFile',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='file_name', full_name='pw.software_update.MetadataFile.file_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='version', full_name='pw.software_update.MetadataFile.version', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='length', full_name='pw.software_update.MetadataFile.length', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='hashes', full_name='pw.software_update.MetadataFile.hashes', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_file_name', full_name='pw.software_update.MetadataFile._file_name',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_length', full_name='pw.software_update.MetadataFile._length',
      index=1, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=2131,
  serialized_end=2274,
)

_SIGNEDROOTMETADATA.fields_by_name['signatures'].message_type = _SIGNATURE
_SIGNEDTIMESTAMPMETADATA.fields_by_name['signatures'].message_type = _SIGNATURE
_SIGNEDSNAPSHOTMETADATA.fields_by_name['signatures'].message_type = _SIGNATURE
_SIGNEDTARGETSMETADATA.fields_by_name['signatures'].message_type = _SIGNATURE
_COMMONMETADATA.fields_by_name['expires'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_COMMONMETADATA.oneofs_by_name['_expires'].fields.append(
  _COMMONMETADATA.fields_by_name['expires'])
_COMMONMETADATA.fields_by_name['expires'].containing_oneof = _COMMONMETADATA.oneofs_by_name['_expires']
_COMMONMETADATA.oneofs_by_name['_role'].fields.append(
  _COMMONMETADATA.fields_by_name['role'])
_COMMONMETADATA.fields_by_name['role'].containing_oneof = _COMMONMETADATA.oneofs_by_name['_role']
_ROOTMETADATA.fields_by_name['common_metadata'].message_type = _COMMONMETADATA
_ROOTMETADATA.fields_by_name['keys'].message_type = _KEYMAPPING
_ROOTMETADATA.fields_by_name['root_signature_requirement'].message_type = _SIGNATUREREQUIREMENT
_ROOTMETADATA.fields_by_name['timestamp_signature_requirement'].message_type = _SIGNATUREREQUIREMENT
_ROOTMETADATA.fields_by_name['snapshot_signature_requirement'].message_type = _SIGNATUREREQUIREMENT
_ROOTMETADATA.fields_by_name['targets_signature_requirement'].message_type = _SIGNATUREREQUIREMENT
_TIMESTAMPMETADATA.fields_by_name['common_metadata'].message_type = _COMMONMETADATA
_TIMESTAMPMETADATA.fields_by_name['snapshot_metadata'].message_type = _METADATAFILE
_SNAPSHOTMETADATA.fields_by_name['common_metadata'].message_type = _COMMONMETADATA
_SNAPSHOTMETADATA.fields_by_name['targets_metadata'].message_type = _METADATAFILE
_TARGETSMETADATA.fields_by_name['common_metadata'].message_type = _COMMONMETADATA
_TARGETSMETADATA.fields_by_name['target_files'].message_type = _TARGETFILE
_KEYMAPPING.fields_by_name['key'].message_type = _KEY
_KEY.fields_by_name['key_type'].enum_type = _KEYTYPE
_KEY.fields_by_name['scheme'].enum_type = _KEYSCHEME
_HASH.fields_by_name['function'].enum_type = _HASHFUNCTION
_TARGETFILE.fields_by_name['hashes'].message_type = _HASH
_METADATAFILE.fields_by_name['hashes'].message_type = _HASH
_METADATAFILE.oneofs_by_name['_file_name'].fields.append(
  _METADATAFILE.fields_by_name['file_name'])
_METADATAFILE.fields_by_name['file_name'].containing_oneof = _METADATAFILE.oneofs_by_name['_file_name']
_METADATAFILE.oneofs_by_name['_length'].fields.append(
  _METADATAFILE.fields_by_name['length'])
_METADATAFILE.fields_by_name['length'].containing_oneof = _METADATAFILE.oneofs_by_name['_length']
DESCRIPTOR.message_types_by_name['SignedRootMetadata'] = _SIGNEDROOTMETADATA
DESCRIPTOR.message_types_by_name['SignedTimestampMetadata'] = _SIGNEDTIMESTAMPMETADATA
DESCRIPTOR.message_types_by_name['SignedSnapshotMetadata'] = _SIGNEDSNAPSHOTMETADATA
DESCRIPTOR.message_types_by_name['SignedTargetsMetadata'] = _SIGNEDTARGETSMETADATA
DESCRIPTOR.message_types_by_name['CommonMetadata'] = _COMMONMETADATA
DESCRIPTOR.message_types_by_name['RootMetadata'] = _ROOTMETADATA
DESCRIPTOR.message_types_by_name['TimestampMetadata'] = _TIMESTAMPMETADATA
DESCRIPTOR.message_types_by_name['SnapshotMetadata'] = _SNAPSHOTMETADATA
DESCRIPTOR.message_types_by_name['TargetsMetadata'] = _TARGETSMETADATA
DESCRIPTOR.message_types_by_name['Signature'] = _SIGNATURE
DESCRIPTOR.message_types_by_name['KeyMapping'] = _KEYMAPPING
DESCRIPTOR.message_types_by_name['Key'] = _KEY
DESCRIPTOR.message_types_by_name['SignatureRequirement'] = _SIGNATUREREQUIREMENT
DESCRIPTOR.message_types_by_name['Hash'] = _HASH
DESCRIPTOR.message_types_by_name['TargetFile'] = _TARGETFILE
DESCRIPTOR.message_types_by_name['MetadataFile'] = _METADATAFILE
DESCRIPTOR.enum_types_by_name['HashFunction'] = _HASHFUNCTION
DESCRIPTOR.enum_types_by_name['KeyType'] = _KEYTYPE
DESCRIPTOR.enum_types_by_name['KeyScheme'] = _KEYSCHEME
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

SignedRootMetadata = _reflection.GeneratedProtocolMessageType('SignedRootMetadata', (_message.Message,), {
  'DESCRIPTOR' : _SIGNEDROOTMETADATA,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.SignedRootMetadata)
  })
_sym_db.RegisterMessage(SignedRootMetadata)

SignedTimestampMetadata = _reflection.GeneratedProtocolMessageType('SignedTimestampMetadata', (_message.Message,), {
  'DESCRIPTOR' : _SIGNEDTIMESTAMPMETADATA,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.SignedTimestampMetadata)
  })
_sym_db.RegisterMessage(SignedTimestampMetadata)

SignedSnapshotMetadata = _reflection.GeneratedProtocolMessageType('SignedSnapshotMetadata', (_message.Message,), {
  'DESCRIPTOR' : _SIGNEDSNAPSHOTMETADATA,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.SignedSnapshotMetadata)
  })
_sym_db.RegisterMessage(SignedSnapshotMetadata)

SignedTargetsMetadata = _reflection.GeneratedProtocolMessageType('SignedTargetsMetadata', (_message.Message,), {
  'DESCRIPTOR' : _SIGNEDTARGETSMETADATA,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.SignedTargetsMetadata)
  })
_sym_db.RegisterMessage(SignedTargetsMetadata)

CommonMetadata = _reflection.GeneratedProtocolMessageType('CommonMetadata', (_message.Message,), {
  'DESCRIPTOR' : _COMMONMETADATA,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.CommonMetadata)
  })
_sym_db.RegisterMessage(CommonMetadata)

RootMetadata = _reflection.GeneratedProtocolMessageType('RootMetadata', (_message.Message,), {
  'DESCRIPTOR' : _ROOTMETADATA,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.RootMetadata)
  })
_sym_db.RegisterMessage(RootMetadata)

TimestampMetadata = _reflection.GeneratedProtocolMessageType('TimestampMetadata', (_message.Message,), {
  'DESCRIPTOR' : _TIMESTAMPMETADATA,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.TimestampMetadata)
  })
_sym_db.RegisterMessage(TimestampMetadata)

SnapshotMetadata = _reflection.GeneratedProtocolMessageType('SnapshotMetadata', (_message.Message,), {
  'DESCRIPTOR' : _SNAPSHOTMETADATA,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.SnapshotMetadata)
  })
_sym_db.RegisterMessage(SnapshotMetadata)

TargetsMetadata = _reflection.GeneratedProtocolMessageType('TargetsMetadata', (_message.Message,), {
  'DESCRIPTOR' : _TARGETSMETADATA,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.TargetsMetadata)
  })
_sym_db.RegisterMessage(TargetsMetadata)

Signature = _reflection.GeneratedProtocolMessageType('Signature', (_message.Message,), {
  'DESCRIPTOR' : _SIGNATURE,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.Signature)
  })
_sym_db.RegisterMessage(Signature)

KeyMapping = _reflection.GeneratedProtocolMessageType('KeyMapping', (_message.Message,), {
  'DESCRIPTOR' : _KEYMAPPING,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.KeyMapping)
  })
_sym_db.RegisterMessage(KeyMapping)

Key = _reflection.GeneratedProtocolMessageType('Key', (_message.Message,), {
  'DESCRIPTOR' : _KEY,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.Key)
  })
_sym_db.RegisterMessage(Key)

SignatureRequirement = _reflection.GeneratedProtocolMessageType('SignatureRequirement', (_message.Message,), {
  'DESCRIPTOR' : _SIGNATUREREQUIREMENT,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.SignatureRequirement)
  })
_sym_db.RegisterMessage(SignatureRequirement)

Hash = _reflection.GeneratedProtocolMessageType('Hash', (_message.Message,), {
  'DESCRIPTOR' : _HASH,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.Hash)
  })
_sym_db.RegisterMessage(Hash)

TargetFile = _reflection.GeneratedProtocolMessageType('TargetFile', (_message.Message,), {
  'DESCRIPTOR' : _TARGETFILE,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.TargetFile)
  })
_sym_db.RegisterMessage(TargetFile)

MetadataFile = _reflection.GeneratedProtocolMessageType('MetadataFile', (_message.Message,), {
  'DESCRIPTOR' : _METADATAFILE,
  '__module__' : 'pw_software_update.tuf_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.MetadataFile)
  })
_sym_db.RegisterMessage(MetadataFile)


# @@protoc_insertion_point(module_scope)
