"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2022 The Pigweed Authors

Licensed under the Apache License, Version 2.0 (the "License"); you may not
use this file except in compliance with the License. You may obtain a copy of
the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations under
the License.
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class CodegenOptions(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MAX_COUNT_FIELD_NUMBER: builtins.int
    FIXED_COUNT_FIELD_NUMBER: builtins.int
    MAX_SIZE_FIELD_NUMBER: builtins.int
    FIXED_SIZE_FIELD_NUMBER: builtins.int
    USE_CALLBACK_FIELD_NUMBER: builtins.int
    max_count: builtins.int
    """Maximum number of entries for repeated fields."""
    fixed_count: builtins.bool
    """Use a fixed length std::array for repeated fields instead of pw::Vector."""
    max_size: builtins.int
    """Maximum size for bytes and string fields."""
    fixed_size: builtins.bool
    """Use a fixed size std::array for bytes fields instead of pw::Vector."""
    use_callback: builtins.bool
    """Force the use of a callback function for the field."""
    def __init__(
        self,
        *,
        max_count: builtins.int = ...,
        fixed_count: builtins.bool = ...,
        max_size: builtins.int = ...,
        fixed_size: builtins.bool = ...,
        use_callback: builtins.bool = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["fixed_count", b"fixed_count", "fixed_size", b"fixed_size", "max_count", b"max_count", "max_size", b"max_size", "use_callback", b"use_callback"]) -> None: ...

global___CodegenOptions = CodegenOptions
