# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_rpc/echo.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_rpc/echo.proto',
  package='pw.rpc',
  syntax='proto3',
  serialized_options=b'\n\030dev.pigweed.pw_rpc.proto',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x11pw_rpc/echo.proto\x12\x06pw.rpc\"\x1a\n\x0b\x45\x63hoMessage\x12\x0b\n\x03msg\x18\x01 \x01(\t2A\n\x0b\x45\x63hoService\x12\x32\n\x04\x45\x63ho\x12\x13.pw.rpc.EchoMessage\x1a\x13.pw.rpc.EchoMessage\"\x00\x42\x1a\n\x18\x64\x65v.pigweed.pw_rpc.protob\x06proto3'
)




_ECHOMESSAGE = _descriptor.Descriptor(
  name='EchoMessage',
  full_name='pw.rpc.EchoMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='msg', full_name='pw.rpc.EchoMessage.msg', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=29,
  serialized_end=55,
)

DESCRIPTOR.message_types_by_name['EchoMessage'] = _ECHOMESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

EchoMessage = _reflection.GeneratedProtocolMessageType('EchoMessage', (_message.Message,), {
  'DESCRIPTOR' : _ECHOMESSAGE,
  '__module__' : 'pw_rpc.echo_pb2'
  # @@protoc_insertion_point(class_scope:pw.rpc.EchoMessage)
  })
_sym_db.RegisterMessage(EchoMessage)


DESCRIPTOR._options = None

_ECHOSERVICE = _descriptor.ServiceDescriptor(
  name='EchoService',
  full_name='pw.rpc.EchoService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=57,
  serialized_end=122,
  methods=[
  _descriptor.MethodDescriptor(
    name='Echo',
    full_name='pw.rpc.EchoService.Echo',
    index=0,
    containing_service=None,
    input_type=_ECHOMESSAGE,
    output_type=_ECHOMESSAGE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_ECHOSERVICE)

DESCRIPTOR.services_by_name['EchoService'] = _ECHOSERVICE

# @@protoc_insertion_point(module_scope)
