# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_software_update/update_bundle.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pw_software_update import tuf_pb2 as pw__software__update_dot_tuf__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_software_update/update_bundle.proto',
  package='pw.software_update',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n&pw_software_update/update_bundle.proto\x12\x12pw.software_update\x1a\x1cpw_software_update/tuf.proto\"\xe5\x04\n\x0cUpdateBundle\x12L\n\x12timestamp_metadata\x18\x01 \x01(\x0b\x32+.pw.software_update.SignedTimestampMetadataH\x00\x88\x01\x01\x12J\n\x11snapshot_metadata\x18\x02 \x01(\x0b\x32*.pw.software_update.SignedSnapshotMetadataH\x01\x88\x01\x01\x12O\n\x10targets_metadata\x18\x03 \x03(\x0b\x32\x35.pw.software_update.UpdateBundle.TargetsMetadataEntry\x12M\n\x0ftarget_payloads\x18\x04 \x03(\x0b\x32\x34.pw.software_update.UpdateBundle.TargetPayloadsEntry\x12\x42\n\rroot_metadata\x18\x05 \x01(\x0b\x32&.pw.software_update.SignedRootMetadataH\x02\x88\x01\x01\x1a\x61\n\x14TargetsMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x38\n\x05value\x18\x02 \x01(\x0b\x32).pw.software_update.SignedTargetsMetadata:\x02\x38\x01\x1a\x35\n\x13TargetPayloadsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c:\x02\x38\x01\x42\x15\n\x13_timestamp_metadataB\x14\n\x12_snapshot_metadataB\x10\n\x0e_root_metadata\"\xe2\x01\n\x08Manifest\x12K\n\x10targets_metadata\x18\x01 \x03(\x0b\x32\x31.pw.software_update.Manifest.TargetsMetadataEntry\x12\x1a\n\ruser_manifest\x18\x02 \x01(\x0cH\x00\x88\x01\x01\x1a[\n\x14TargetsMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x32\n\x05value\x18\x02 \x01(\x0b\x32#.pw.software_update.TargetsMetadata:\x02\x38\x01\x42\x10\n\x0e_user_manifestb\x06proto3'
  ,
  dependencies=[pw__software__update_dot_tuf__pb2.DESCRIPTOR,])




_UPDATEBUNDLE_TARGETSMETADATAENTRY = _descriptor.Descriptor(
  name='TargetsMetadataEntry',
  full_name='pw.software_update.UpdateBundle.TargetsMetadataEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='pw.software_update.UpdateBundle.TargetsMetadataEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='pw.software_update.UpdateBundle.TargetsMetadataEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=491,
  serialized_end=588,
)

_UPDATEBUNDLE_TARGETPAYLOADSENTRY = _descriptor.Descriptor(
  name='TargetPayloadsEntry',
  full_name='pw.software_update.UpdateBundle.TargetPayloadsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='pw.software_update.UpdateBundle.TargetPayloadsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='pw.software_update.UpdateBundle.TargetPayloadsEntry.value', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=590,
  serialized_end=643,
)

_UPDATEBUNDLE = _descriptor.Descriptor(
  name='UpdateBundle',
  full_name='pw.software_update.UpdateBundle',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='timestamp_metadata', full_name='pw.software_update.UpdateBundle.timestamp_metadata', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='snapshot_metadata', full_name='pw.software_update.UpdateBundle.snapshot_metadata', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='targets_metadata', full_name='pw.software_update.UpdateBundle.targets_metadata', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='target_payloads', full_name='pw.software_update.UpdateBundle.target_payloads', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='root_metadata', full_name='pw.software_update.UpdateBundle.root_metadata', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_UPDATEBUNDLE_TARGETSMETADATAENTRY, _UPDATEBUNDLE_TARGETPAYLOADSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_timestamp_metadata', full_name='pw.software_update.UpdateBundle._timestamp_metadata',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_snapshot_metadata', full_name='pw.software_update.UpdateBundle._snapshot_metadata',
      index=1, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_root_metadata', full_name='pw.software_update.UpdateBundle._root_metadata',
      index=2, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=93,
  serialized_end=706,
)


_MANIFEST_TARGETSMETADATAENTRY = _descriptor.Descriptor(
  name='TargetsMetadataEntry',
  full_name='pw.software_update.Manifest.TargetsMetadataEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='pw.software_update.Manifest.TargetsMetadataEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='pw.software_update.Manifest.TargetsMetadataEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=826,
  serialized_end=917,
)

_MANIFEST = _descriptor.Descriptor(
  name='Manifest',
  full_name='pw.software_update.Manifest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='targets_metadata', full_name='pw.software_update.Manifest.targets_metadata', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_manifest', full_name='pw.software_update.Manifest.user_manifest', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_MANIFEST_TARGETSMETADATAENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_user_manifest', full_name='pw.software_update.Manifest._user_manifest',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=709,
  serialized_end=935,
)

_UPDATEBUNDLE_TARGETSMETADATAENTRY.fields_by_name['value'].message_type = pw__software__update_dot_tuf__pb2._SIGNEDTARGETSMETADATA
_UPDATEBUNDLE_TARGETSMETADATAENTRY.containing_type = _UPDATEBUNDLE
_UPDATEBUNDLE_TARGETPAYLOADSENTRY.containing_type = _UPDATEBUNDLE
_UPDATEBUNDLE.fields_by_name['timestamp_metadata'].message_type = pw__software__update_dot_tuf__pb2._SIGNEDTIMESTAMPMETADATA
_UPDATEBUNDLE.fields_by_name['snapshot_metadata'].message_type = pw__software__update_dot_tuf__pb2._SIGNEDSNAPSHOTMETADATA
_UPDATEBUNDLE.fields_by_name['targets_metadata'].message_type = _UPDATEBUNDLE_TARGETSMETADATAENTRY
_UPDATEBUNDLE.fields_by_name['target_payloads'].message_type = _UPDATEBUNDLE_TARGETPAYLOADSENTRY
_UPDATEBUNDLE.fields_by_name['root_metadata'].message_type = pw__software__update_dot_tuf__pb2._SIGNEDROOTMETADATA
_UPDATEBUNDLE.oneofs_by_name['_timestamp_metadata'].fields.append(
  _UPDATEBUNDLE.fields_by_name['timestamp_metadata'])
_UPDATEBUNDLE.fields_by_name['timestamp_metadata'].containing_oneof = _UPDATEBUNDLE.oneofs_by_name['_timestamp_metadata']
_UPDATEBUNDLE.oneofs_by_name['_snapshot_metadata'].fields.append(
  _UPDATEBUNDLE.fields_by_name['snapshot_metadata'])
_UPDATEBUNDLE.fields_by_name['snapshot_metadata'].containing_oneof = _UPDATEBUNDLE.oneofs_by_name['_snapshot_metadata']
_UPDATEBUNDLE.oneofs_by_name['_root_metadata'].fields.append(
  _UPDATEBUNDLE.fields_by_name['root_metadata'])
_UPDATEBUNDLE.fields_by_name['root_metadata'].containing_oneof = _UPDATEBUNDLE.oneofs_by_name['_root_metadata']
_MANIFEST_TARGETSMETADATAENTRY.fields_by_name['value'].message_type = pw__software__update_dot_tuf__pb2._TARGETSMETADATA
_MANIFEST_TARGETSMETADATAENTRY.containing_type = _MANIFEST
_MANIFEST.fields_by_name['targets_metadata'].message_type = _MANIFEST_TARGETSMETADATAENTRY
_MANIFEST.oneofs_by_name['_user_manifest'].fields.append(
  _MANIFEST.fields_by_name['user_manifest'])
_MANIFEST.fields_by_name['user_manifest'].containing_oneof = _MANIFEST.oneofs_by_name['_user_manifest']
DESCRIPTOR.message_types_by_name['UpdateBundle'] = _UPDATEBUNDLE
DESCRIPTOR.message_types_by_name['Manifest'] = _MANIFEST
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

UpdateBundle = _reflection.GeneratedProtocolMessageType('UpdateBundle', (_message.Message,), {

  'TargetsMetadataEntry' : _reflection.GeneratedProtocolMessageType('TargetsMetadataEntry', (_message.Message,), {
    'DESCRIPTOR' : _UPDATEBUNDLE_TARGETSMETADATAENTRY,
    '__module__' : 'pw_software_update.update_bundle_pb2'
    # @@protoc_insertion_point(class_scope:pw.software_update.UpdateBundle.TargetsMetadataEntry)
    })
  ,

  'TargetPayloadsEntry' : _reflection.GeneratedProtocolMessageType('TargetPayloadsEntry', (_message.Message,), {
    'DESCRIPTOR' : _UPDATEBUNDLE_TARGETPAYLOADSENTRY,
    '__module__' : 'pw_software_update.update_bundle_pb2'
    # @@protoc_insertion_point(class_scope:pw.software_update.UpdateBundle.TargetPayloadsEntry)
    })
  ,
  'DESCRIPTOR' : _UPDATEBUNDLE,
  '__module__' : 'pw_software_update.update_bundle_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.UpdateBundle)
  })
_sym_db.RegisterMessage(UpdateBundle)
_sym_db.RegisterMessage(UpdateBundle.TargetsMetadataEntry)
_sym_db.RegisterMessage(UpdateBundle.TargetPayloadsEntry)

Manifest = _reflection.GeneratedProtocolMessageType('Manifest', (_message.Message,), {

  'TargetsMetadataEntry' : _reflection.GeneratedProtocolMessageType('TargetsMetadataEntry', (_message.Message,), {
    'DESCRIPTOR' : _MANIFEST_TARGETSMETADATAENTRY,
    '__module__' : 'pw_software_update.update_bundle_pb2'
    # @@protoc_insertion_point(class_scope:pw.software_update.Manifest.TargetsMetadataEntry)
    })
  ,
  'DESCRIPTOR' : _MANIFEST,
  '__module__' : 'pw_software_update.update_bundle_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.Manifest)
  })
_sym_db.RegisterMessage(Manifest)
_sym_db.RegisterMessage(Manifest.TargetsMetadataEntry)


_UPDATEBUNDLE_TARGETSMETADATAENTRY._options = None
_UPDATEBUNDLE_TARGETPAYLOADSENTRY._options = None
_MANIFEST_TARGETSMETADATAENTRY._options = None
# @@protoc_insertion_point(module_scope)
