# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_thread_protos/thread.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pw_tokenizer.proto import options_pb2 as pw__tokenizer_dot_proto_dot_options__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_thread_protos/thread.proto',
  package='pw.thread.proto',
  syntax='proto3',
  serialized_options=b'\n\017pw.thread.protoB\006Thread',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1dpw_thread_protos/thread.proto\x12\x0fpw.thread.proto\x1a pw_tokenizer/proto/options.proto\"{\n\x0bThreadState\"l\n\x04\x45num\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x15\n\x11INTERRUPT_HANDLER\x10\x01\x12\x0b\n\x07RUNNING\x10\x02\x12\t\n\x05READY\x10\x03\x12\r\n\tSUSPENDED\x10\x04\x12\x0b\n\x07\x42LOCKED\x10\x05\x12\x0c\n\x08INACTIVE\x10\x06\"\xc4\x03\n\x06Thread\x12\x12\n\x04name\x18\x01 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\x12\n\x06\x61\x63tive\x18\x02 \x01(\x08\x42\x02\x18\x01\x12\x30\n\x05state\x18\x03 \x01(\x0e\x32!.pw.thread.proto.ThreadState.Enum\x12\x15\n\rraw_backtrace\x18\x04 \x03(\x04\x12\x1c\n\x14symbolized_backtrace\x18\x05 \x03(\t\x12\x11\n\traw_stack\x18\x06 \x01(\x0c\x12 \n\x13stack_start_pointer\x18\x07 \x01(\x04H\x00\x88\x01\x01\x12\x1e\n\x11stack_end_pointer\x18\x08 \x01(\x04H\x01\x88\x01\x01\x12\x1a\n\rstack_pointer\x18\t \x01(\x04H\x02\x88\x01\x01\x12!\n\x14\x63pu_usage_hundredths\x18\n \x01(\rH\x03\x88\x01\x01\x12#\n\x16stack_pointer_est_peak\x18\x0b \x01(\x04H\x04\x88\x01\x01\x42\x16\n\x14_stack_start_pointerB\x14\n\x12_stack_end_pointerB\x10\n\x0e_stack_pointerB\x17\n\x15_cpu_usage_hundredthsB\x19\n\x17_stack_pointer_est_peak\">\n\x12SnapshotThreadInfo\x12(\n\x07threads\x18\x12 \x03(\x0b\x32\x17.pw.thread.proto.ThreadB\x19\n\x0fpw.thread.protoB\x06Threadb\x06proto3'
  ,
  dependencies=[pw__tokenizer_dot_proto_dot_options__pb2.DESCRIPTOR,])



_THREADSTATE_ENUM = _descriptor.EnumDescriptor(
  name='Enum',
  full_name='pw.thread.proto.ThreadState.Enum',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INTERRUPT_HANDLER', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='RUNNING', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='READY', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SUSPENDED', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BLOCKED', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INACTIVE', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=99,
  serialized_end=207,
)
_sym_db.RegisterEnumDescriptor(_THREADSTATE_ENUM)


_THREADSTATE = _descriptor.Descriptor(
  name='ThreadState',
  full_name='pw.thread.proto.ThreadState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _THREADSTATE_ENUM,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=84,
  serialized_end=207,
)


_THREAD = _descriptor.Descriptor(
  name='Thread',
  full_name='pw.thread.proto.Thread',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='pw.thread.proto.Thread.name', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\200\257&\000', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='active', full_name='pw.thread.proto.Thread.active', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\030\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='pw.thread.proto.Thread.state', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='raw_backtrace', full_name='pw.thread.proto.Thread.raw_backtrace', index=3,
      number=4, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='symbolized_backtrace', full_name='pw.thread.proto.Thread.symbolized_backtrace', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='raw_stack', full_name='pw.thread.proto.Thread.raw_stack', index=5,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='stack_start_pointer', full_name='pw.thread.proto.Thread.stack_start_pointer', index=6,
      number=7, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='stack_end_pointer', full_name='pw.thread.proto.Thread.stack_end_pointer', index=7,
      number=8, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='stack_pointer', full_name='pw.thread.proto.Thread.stack_pointer', index=8,
      number=9, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cpu_usage_hundredths', full_name='pw.thread.proto.Thread.cpu_usage_hundredths', index=9,
      number=10, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='stack_pointer_est_peak', full_name='pw.thread.proto.Thread.stack_pointer_est_peak', index=10,
      number=11, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_stack_start_pointer', full_name='pw.thread.proto.Thread._stack_start_pointer',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_stack_end_pointer', full_name='pw.thread.proto.Thread._stack_end_pointer',
      index=1, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_stack_pointer', full_name='pw.thread.proto.Thread._stack_pointer',
      index=2, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_cpu_usage_hundredths', full_name='pw.thread.proto.Thread._cpu_usage_hundredths',
      index=3, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_stack_pointer_est_peak', full_name='pw.thread.proto.Thread._stack_pointer_est_peak',
      index=4, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=210,
  serialized_end=662,
)


_SNAPSHOTTHREADINFO = _descriptor.Descriptor(
  name='SnapshotThreadInfo',
  full_name='pw.thread.proto.SnapshotThreadInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='threads', full_name='pw.thread.proto.SnapshotThreadInfo.threads', index=0,
      number=18, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=664,
  serialized_end=726,
)

_THREADSTATE_ENUM.containing_type = _THREADSTATE
_THREAD.fields_by_name['state'].enum_type = _THREADSTATE_ENUM
_THREAD.oneofs_by_name['_stack_start_pointer'].fields.append(
  _THREAD.fields_by_name['stack_start_pointer'])
_THREAD.fields_by_name['stack_start_pointer'].containing_oneof = _THREAD.oneofs_by_name['_stack_start_pointer']
_THREAD.oneofs_by_name['_stack_end_pointer'].fields.append(
  _THREAD.fields_by_name['stack_end_pointer'])
_THREAD.fields_by_name['stack_end_pointer'].containing_oneof = _THREAD.oneofs_by_name['_stack_end_pointer']
_THREAD.oneofs_by_name['_stack_pointer'].fields.append(
  _THREAD.fields_by_name['stack_pointer'])
_THREAD.fields_by_name['stack_pointer'].containing_oneof = _THREAD.oneofs_by_name['_stack_pointer']
_THREAD.oneofs_by_name['_cpu_usage_hundredths'].fields.append(
  _THREAD.fields_by_name['cpu_usage_hundredths'])
_THREAD.fields_by_name['cpu_usage_hundredths'].containing_oneof = _THREAD.oneofs_by_name['_cpu_usage_hundredths']
_THREAD.oneofs_by_name['_stack_pointer_est_peak'].fields.append(
  _THREAD.fields_by_name['stack_pointer_est_peak'])
_THREAD.fields_by_name['stack_pointer_est_peak'].containing_oneof = _THREAD.oneofs_by_name['_stack_pointer_est_peak']
_SNAPSHOTTHREADINFO.fields_by_name['threads'].message_type = _THREAD
DESCRIPTOR.message_types_by_name['ThreadState'] = _THREADSTATE
DESCRIPTOR.message_types_by_name['Thread'] = _THREAD
DESCRIPTOR.message_types_by_name['SnapshotThreadInfo'] = _SNAPSHOTTHREADINFO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ThreadState = _reflection.GeneratedProtocolMessageType('ThreadState', (_message.Message,), {
  'DESCRIPTOR' : _THREADSTATE,
  '__module__' : 'pw_thread_protos.thread_pb2'
  # @@protoc_insertion_point(class_scope:pw.thread.proto.ThreadState)
  })
_sym_db.RegisterMessage(ThreadState)

Thread = _reflection.GeneratedProtocolMessageType('Thread', (_message.Message,), {
  'DESCRIPTOR' : _THREAD,
  '__module__' : 'pw_thread_protos.thread_pb2'
  # @@protoc_insertion_point(class_scope:pw.thread.proto.Thread)
  })
_sym_db.RegisterMessage(Thread)

SnapshotThreadInfo = _reflection.GeneratedProtocolMessageType('SnapshotThreadInfo', (_message.Message,), {
  'DESCRIPTOR' : _SNAPSHOTTHREADINFO,
  '__module__' : 'pw_thread_protos.thread_pb2'
  # @@protoc_insertion_point(class_scope:pw.thread.proto.SnapshotThreadInfo)
  })
_sym_db.RegisterMessage(SnapshotThreadInfo)


DESCRIPTOR._options = None
_THREAD.fields_by_name['name']._options = None
_THREAD.fields_by_name['active']._options = None
# @@protoc_insertion_point(module_scope)
