# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Defines argument types for use with argparse."""

import argparse
import logging
from pathlib import Path


def directory(arg: str) -> Path:
    path = Path(arg)
    if path.is_dir():
        return path.resolve()

    raise argparse.ArgumentTypeError(f'"{path}" is not a directory')


def log_level(arg: str) -> int:
    try:
        return getattr(logging, arg.upper())
    except AttributeError:
        raise argparse.ArgumentTypeError(
            f'"{arg.upper()}" is not a valid log level'
        )
