# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_snapshot_protos/snapshot.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pw_chrono_protos import chrono_pb2 as pw__chrono__protos_dot_chrono__pb2
from pw_cpu_exception_cortex_m_protos import cpu_state_pb2 as pw__cpu__exception__cortex__m__protos_dot_cpu__state__pb2
from pw_log.proto import log_pb2 as pw__log_dot_proto_dot_log__pb2
from pw_thread_protos import thread_pb2 as pw__thread__protos_dot_thread__pb2
from pw_snapshot_metadata_proto import snapshot_metadata_pb2 as pw__snapshot__metadata__proto_dot_snapshot__metadata__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_snapshot_protos/snapshot.proto',
  package='pw.snapshot',
  syntax='proto3',
  serialized_options=b'\n\021pw.snapshot.protoB\010Snapshot',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n!pw_snapshot_protos/snapshot.proto\x12\x0bpw.snapshot\x1a\x1dpw_chrono_protos/chrono.proto\x1a\x30pw_cpu_exception_cortex_m_protos/cpu_state.proto\x1a\x16pw_log/proto/log.proto\x1a\x1dpw_thread_protos/thread.proto\x1a\x32pw_snapshot_metadata_proto/snapshot_metadata.proto\"\xac\x03\n\x08Snapshot\x12\x1e\n\x04logs\x18\x01 \x03(\x0b\x32\x10.pw.log.LogEntry\x12\'\n\x08metadata\x18\x10 \x01(\x0b\x32\x15.pw.snapshot.Metadata\x12-\n\x04tags\x18\x11 \x03(\x0b\x32\x1f.pw.snapshot.Snapshot.TagsEntry\x12(\n\x07threads\x18\x12 \x03(\x0b\x32\x17.pw.thread.proto.Thread\x12\x30\n\x11related_snapshots\x18\x13 \x03(\x0b\x32\x15.pw.snapshot.Snapshot\x12\x43\n\x10\x61rmv7m_cpu_state\x18\x14 \x01(\x0b\x32).pw.cpu_exception.cortex_m.ArmV7mCpuState\x12\x12\n\ntrace_data\x18\x15 \x01(\x0c\x12(\n\ntimestamps\x18\x16 \x03(\x0b\x32\x14.pw.chrono.TimePoint\x1a+\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01J\x04\x08\x02\x10\x08J\x04\x08\x08\x10\x10J\x05\x08\x17\x10\x88\x08J\t\x08\x88\x08\x10\x80\x80\x80\x80\x02\x42\x1d\n\x11pw.snapshot.protoB\x08Snapshotb\x06proto3'
  ,
  dependencies=[pw__chrono__protos_dot_chrono__pb2.DESCRIPTOR,pw__cpu__exception__cortex__m__protos_dot_cpu__state__pb2.DESCRIPTOR,pw__log_dot_proto_dot_log__pb2.DESCRIPTOR,pw__thread__protos_dot_thread__pb2.DESCRIPTOR,pw__snapshot__metadata__proto_dot_snapshot__metadata__pb2.DESCRIPTOR,])




_SNAPSHOT_TAGSENTRY = _descriptor.Descriptor(
  name='TagsEntry',
  full_name='pw.snapshot.Snapshot.TagsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='pw.snapshot.Snapshot.TagsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='pw.snapshot.Snapshot.TagsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=594,
  serialized_end=637,
)

_SNAPSHOT = _descriptor.Descriptor(
  name='Snapshot',
  full_name='pw.snapshot.Snapshot',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='logs', full_name='pw.snapshot.Snapshot.logs', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metadata', full_name='pw.snapshot.Snapshot.metadata', index=1,
      number=16, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tags', full_name='pw.snapshot.Snapshot.tags', index=2,
      number=17, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='threads', full_name='pw.snapshot.Snapshot.threads', index=3,
      number=18, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='related_snapshots', full_name='pw.snapshot.Snapshot.related_snapshots', index=4,
      number=19, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='armv7m_cpu_state', full_name='pw.snapshot.Snapshot.armv7m_cpu_state', index=5,
      number=20, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='trace_data', full_name='pw.snapshot.Snapshot.trace_data', index=6,
      number=21, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamps', full_name='pw.snapshot.Snapshot.timestamps', index=7,
      number=22, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_SNAPSHOT_TAGSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=239,
  serialized_end=667,
)

_SNAPSHOT_TAGSENTRY.containing_type = _SNAPSHOT
_SNAPSHOT.fields_by_name['logs'].message_type = pw__log_dot_proto_dot_log__pb2._LOGENTRY
_SNAPSHOT.fields_by_name['metadata'].message_type = pw__snapshot__metadata__proto_dot_snapshot__metadata__pb2._METADATA
_SNAPSHOT.fields_by_name['tags'].message_type = _SNAPSHOT_TAGSENTRY
_SNAPSHOT.fields_by_name['threads'].message_type = pw__thread__protos_dot_thread__pb2._THREAD
_SNAPSHOT.fields_by_name['related_snapshots'].message_type = _SNAPSHOT
_SNAPSHOT.fields_by_name['armv7m_cpu_state'].message_type = pw__cpu__exception__cortex__m__protos_dot_cpu__state__pb2._ARMV7MCPUSTATE
_SNAPSHOT.fields_by_name['timestamps'].message_type = pw__chrono__protos_dot_chrono__pb2._TIMEPOINT
DESCRIPTOR.message_types_by_name['Snapshot'] = _SNAPSHOT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Snapshot = _reflection.GeneratedProtocolMessageType('Snapshot', (_message.Message,), {

  'TagsEntry' : _reflection.GeneratedProtocolMessageType('TagsEntry', (_message.Message,), {
    'DESCRIPTOR' : _SNAPSHOT_TAGSENTRY,
    '__module__' : 'pw_snapshot_protos.snapshot_pb2'
    # @@protoc_insertion_point(class_scope:pw.snapshot.Snapshot.TagsEntry)
    })
  ,
  'DESCRIPTOR' : _SNAPSHOT,
  '__module__' : 'pw_snapshot_protos.snapshot_pb2'
  # @@protoc_insertion_point(class_scope:pw.snapshot.Snapshot)
  })
_sym_db.RegisterMessage(Snapshot)
_sym_db.RegisterMessage(Snapshot.TagsEntry)


DESCRIPTOR._options = None
_SNAPSHOT_TAGSENTRY._options = None
# @@protoc_insertion_point(module_scope)
