"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2021 The Pigweed Authors

Licensed under the Apache License, Version 2.0 (the "License"); you may not
use this file except in compliance with the License. You may obtain a copy of
the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations under
the License.
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.descriptor_pb2
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.internal.extension_dict
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _Tokenization:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _TokenizationEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_Tokenization.ValueType], builtins.type):  # noqa: F821
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    TOKENIZATION_OPTIONAL: _Tokenization.ValueType  # 0
    """The field may contain plain text or any type of tokenized data (binary or
    prefixed Base64).
    """

class Tokenization(_Tokenization, metaclass=_TokenizationEnumTypeWrapper): ...

TOKENIZATION_OPTIONAL: Tokenization.ValueType  # 0
"""The field may contain plain text or any type of tokenized data (binary or
prefixed Base64).
"""
global___Tokenization = Tokenization

FORMAT_FIELD_NUMBER: builtins.int
format: google.protobuf.internal.extension_dict._ExtensionFieldDescriptor[google.protobuf.descriptor_pb2.FieldOptions, global___Tokenization.ValueType]
"""The field number was randomly selected from the reserved, internal use
field numbers (50000-99999).
TODO(b/234886465): Register with the Protobuf Global Extension Registry:
    https://github.com/protocolbuffers/protobuf/blob/HEAD/docs/options.md
"""
