# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Outputs a ReST message indicating that size reports don't work on Windows."""

import os
import sys

_NO_BLOATY_ERROR: str = '''
.. note::

  The size reporting tool is not yet available on Windows systems.
'''


def main() -> int:
    os.makedirs(os.path.dirname(sys.argv[1]), exist_ok=True)
    with open(sys.argv[1], 'w') as fd:
        fd.write(_NO_BLOATY_ERROR)
    return 0


if __name__ == '__main__':
    sys.exit(main())
