# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_unit_test_proto/unit_test.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_unit_test_proto/unit_test.proto',
  package='pw.unit_test',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\"pw_unit_test_proto/unit_test.proto\x12\x0cpw.unit_test\"N\n\x12TestCaseDescriptor\x12\x12\n\nsuite_name\x18\x01 \x01(\t\x12\x11\n\ttest_name\x18\x02 \x01(\t\x12\x11\n\tfile_name\x18\x03 \x01(\t\"m\n\x13TestCaseExpectation\x12\x12\n\nexpression\x18\x01 \x01(\t\x12\x1c\n\x14\x65valuated_expression\x18\x02 \x01(\t\x12\x13\n\x0bline_number\x18\x03 \x01(\r\x12\x0f\n\x07success\x18\x04 \x01(\x08\"\x0e\n\x0cTestRunStart\"O\n\nTestRunEnd\x12\x0e\n\x06passed\x18\x01 \x01(\r\x12\x0e\n\x06\x66\x61iled\x18\x02 \x01(\r\x12\x0f\n\x07skipped\x18\x03 \x01(\r\x12\x10\n\x08\x64isabled\x18\x04 \x01(\r\"\xef\x02\n\x05\x45vent\x12\x34\n\x0etest_run_start\x18\x01 \x01(\x0b\x32\x1a.pw.unit_test.TestRunStartH\x00\x12\x30\n\x0ctest_run_end\x18\x02 \x01(\x0b\x32\x18.pw.unit_test.TestRunEndH\x00\x12;\n\x0ftest_case_start\x18\x03 \x01(\x0b\x32 .pw.unit_test.TestCaseDescriptorH\x00\x12\x35\n\rtest_case_end\x18\x04 \x01(\x0e\x32\x1c.pw.unit_test.TestCaseResultH\x00\x12>\n\x12test_case_disabled\x18\x05 \x01(\x0b\x32 .pw.unit_test.TestCaseDescriptorH\x00\x12\x42\n\x15test_case_expectation\x18\x06 \x01(\x0b\x32!.pw.unit_test.TestCaseExpectationH\x00\x42\x06\n\x04type\"H\n\x0eTestRunRequest\x12\"\n\x1areport_passed_expectations\x18\x01 \x01(\x08\x12\x12\n\ntest_suite\x18\x02 \x03(\t*7\n\x0eTestCaseResult\x12\x0b\n\x07SUCCESS\x10\x00\x12\x0b\n\x07\x46\x41ILURE\x10\x01\x12\x0b\n\x07SKIPPED\x10\x02\x32H\n\x08UnitTest\x12<\n\x03Run\x12\x1c.pw.unit_test.TestRunRequest\x1a\x13.pw.unit_test.Event\"\x00\x30\x01\x62\x06proto3'
)

_TESTCASERESULT = _descriptor.EnumDescriptor(
  name='TestCaseResult',
  full_name='pw.unit_test.TestCaseResult',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SUCCESS', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FAILURE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SKIPPED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=784,
  serialized_end=839,
)
_sym_db.RegisterEnumDescriptor(_TESTCASERESULT)

TestCaseResult = enum_type_wrapper.EnumTypeWrapper(_TESTCASERESULT)
SUCCESS = 0
FAILURE = 1
SKIPPED = 2



_TESTCASEDESCRIPTOR = _descriptor.Descriptor(
  name='TestCaseDescriptor',
  full_name='pw.unit_test.TestCaseDescriptor',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='suite_name', full_name='pw.unit_test.TestCaseDescriptor.suite_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='test_name', full_name='pw.unit_test.TestCaseDescriptor.test_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='file_name', full_name='pw.unit_test.TestCaseDescriptor.file_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=52,
  serialized_end=130,
)


_TESTCASEEXPECTATION = _descriptor.Descriptor(
  name='TestCaseExpectation',
  full_name='pw.unit_test.TestCaseExpectation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='expression', full_name='pw.unit_test.TestCaseExpectation.expression', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='evaluated_expression', full_name='pw.unit_test.TestCaseExpectation.evaluated_expression', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='line_number', full_name='pw.unit_test.TestCaseExpectation.line_number', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='success', full_name='pw.unit_test.TestCaseExpectation.success', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=132,
  serialized_end=241,
)


_TESTRUNSTART = _descriptor.Descriptor(
  name='TestRunStart',
  full_name='pw.unit_test.TestRunStart',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=243,
  serialized_end=257,
)


_TESTRUNEND = _descriptor.Descriptor(
  name='TestRunEnd',
  full_name='pw.unit_test.TestRunEnd',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='passed', full_name='pw.unit_test.TestRunEnd.passed', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='failed', full_name='pw.unit_test.TestRunEnd.failed', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='skipped', full_name='pw.unit_test.TestRunEnd.skipped', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='disabled', full_name='pw.unit_test.TestRunEnd.disabled', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=259,
  serialized_end=338,
)


_EVENT = _descriptor.Descriptor(
  name='Event',
  full_name='pw.unit_test.Event',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='test_run_start', full_name='pw.unit_test.Event.test_run_start', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='test_run_end', full_name='pw.unit_test.Event.test_run_end', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='test_case_start', full_name='pw.unit_test.Event.test_case_start', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='test_case_end', full_name='pw.unit_test.Event.test_case_end', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='test_case_disabled', full_name='pw.unit_test.Event.test_case_disabled', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='test_case_expectation', full_name='pw.unit_test.Event.test_case_expectation', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='type', full_name='pw.unit_test.Event.type',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=341,
  serialized_end=708,
)


_TESTRUNREQUEST = _descriptor.Descriptor(
  name='TestRunRequest',
  full_name='pw.unit_test.TestRunRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='report_passed_expectations', full_name='pw.unit_test.TestRunRequest.report_passed_expectations', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='test_suite', full_name='pw.unit_test.TestRunRequest.test_suite', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=710,
  serialized_end=782,
)

_EVENT.fields_by_name['test_run_start'].message_type = _TESTRUNSTART
_EVENT.fields_by_name['test_run_end'].message_type = _TESTRUNEND
_EVENT.fields_by_name['test_case_start'].message_type = _TESTCASEDESCRIPTOR
_EVENT.fields_by_name['test_case_end'].enum_type = _TESTCASERESULT
_EVENT.fields_by_name['test_case_disabled'].message_type = _TESTCASEDESCRIPTOR
_EVENT.fields_by_name['test_case_expectation'].message_type = _TESTCASEEXPECTATION
_EVENT.oneofs_by_name['type'].fields.append(
  _EVENT.fields_by_name['test_run_start'])
_EVENT.fields_by_name['test_run_start'].containing_oneof = _EVENT.oneofs_by_name['type']
_EVENT.oneofs_by_name['type'].fields.append(
  _EVENT.fields_by_name['test_run_end'])
_EVENT.fields_by_name['test_run_end'].containing_oneof = _EVENT.oneofs_by_name['type']
_EVENT.oneofs_by_name['type'].fields.append(
  _EVENT.fields_by_name['test_case_start'])
_EVENT.fields_by_name['test_case_start'].containing_oneof = _EVENT.oneofs_by_name['type']
_EVENT.oneofs_by_name['type'].fields.append(
  _EVENT.fields_by_name['test_case_end'])
_EVENT.fields_by_name['test_case_end'].containing_oneof = _EVENT.oneofs_by_name['type']
_EVENT.oneofs_by_name['type'].fields.append(
  _EVENT.fields_by_name['test_case_disabled'])
_EVENT.fields_by_name['test_case_disabled'].containing_oneof = _EVENT.oneofs_by_name['type']
_EVENT.oneofs_by_name['type'].fields.append(
  _EVENT.fields_by_name['test_case_expectation'])
_EVENT.fields_by_name['test_case_expectation'].containing_oneof = _EVENT.oneofs_by_name['type']
DESCRIPTOR.message_types_by_name['TestCaseDescriptor'] = _TESTCASEDESCRIPTOR
DESCRIPTOR.message_types_by_name['TestCaseExpectation'] = _TESTCASEEXPECTATION
DESCRIPTOR.message_types_by_name['TestRunStart'] = _TESTRUNSTART
DESCRIPTOR.message_types_by_name['TestRunEnd'] = _TESTRUNEND
DESCRIPTOR.message_types_by_name['Event'] = _EVENT
DESCRIPTOR.message_types_by_name['TestRunRequest'] = _TESTRUNREQUEST
DESCRIPTOR.enum_types_by_name['TestCaseResult'] = _TESTCASERESULT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

TestCaseDescriptor = _reflection.GeneratedProtocolMessageType('TestCaseDescriptor', (_message.Message,), {
  'DESCRIPTOR' : _TESTCASEDESCRIPTOR,
  '__module__' : 'pw_unit_test_proto.unit_test_pb2'
  # @@protoc_insertion_point(class_scope:pw.unit_test.TestCaseDescriptor)
  })
_sym_db.RegisterMessage(TestCaseDescriptor)

TestCaseExpectation = _reflection.GeneratedProtocolMessageType('TestCaseExpectation', (_message.Message,), {
  'DESCRIPTOR' : _TESTCASEEXPECTATION,
  '__module__' : 'pw_unit_test_proto.unit_test_pb2'
  # @@protoc_insertion_point(class_scope:pw.unit_test.TestCaseExpectation)
  })
_sym_db.RegisterMessage(TestCaseExpectation)

TestRunStart = _reflection.GeneratedProtocolMessageType('TestRunStart', (_message.Message,), {
  'DESCRIPTOR' : _TESTRUNSTART,
  '__module__' : 'pw_unit_test_proto.unit_test_pb2'
  # @@protoc_insertion_point(class_scope:pw.unit_test.TestRunStart)
  })
_sym_db.RegisterMessage(TestRunStart)

TestRunEnd = _reflection.GeneratedProtocolMessageType('TestRunEnd', (_message.Message,), {
  'DESCRIPTOR' : _TESTRUNEND,
  '__module__' : 'pw_unit_test_proto.unit_test_pb2'
  # @@protoc_insertion_point(class_scope:pw.unit_test.TestRunEnd)
  })
_sym_db.RegisterMessage(TestRunEnd)

Event = _reflection.GeneratedProtocolMessageType('Event', (_message.Message,), {
  'DESCRIPTOR' : _EVENT,
  '__module__' : 'pw_unit_test_proto.unit_test_pb2'
  # @@protoc_insertion_point(class_scope:pw.unit_test.Event)
  })
_sym_db.RegisterMessage(Event)

TestRunRequest = _reflection.GeneratedProtocolMessageType('TestRunRequest', (_message.Message,), {
  'DESCRIPTOR' : _TESTRUNREQUEST,
  '__module__' : 'pw_unit_test_proto.unit_test_pb2'
  # @@protoc_insertion_point(class_scope:pw.unit_test.TestRunRequest)
  })
_sym_db.RegisterMessage(TestRunRequest)



_UNITTEST = _descriptor.ServiceDescriptor(
  name='UnitTest',
  full_name='pw.unit_test.UnitTest',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=841,
  serialized_end=913,
  methods=[
  _descriptor.MethodDescriptor(
    name='Run',
    full_name='pw.unit_test.UnitTest.Run',
    index=0,
    containing_service=None,
    input_type=_TESTRUNREQUEST,
    output_type=_EVENT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_UNITTEST)

DESCRIPTOR.services_by_name['UnitTest'] = _UNITTEST

# @@protoc_insertion_point(module_scope)
