# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_thread_protos/thread_snapshot_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pw_thread_protos import thread_pb2 as pw__thread__protos_dot_thread__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_thread_protos/thread_snapshot_service.proto',
  package='pw.thread.proto',
  syntax='proto3',
  serialized_options=b'\n\017pw.thread.protoB\006Thread',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n.pw_thread_protos/thread_snapshot_service.proto\x12\x0fpw.thread.proto\x1a\x1dpw_thread_protos/thread.proto\"+\n\rThreadRequest\x12\x11\n\x04name\x18\x01 \x01(\x0cH\x00\x88\x01\x01\x42\x07\n\x05_name2u\n\x15ThreadSnapshotService\x12\\\n\x11GetPeakStackUsage\x12\x1e.pw.thread.proto.ThreadRequest\x1a#.pw.thread.proto.SnapshotThreadInfo\"\x00\x30\x01\x42\x19\n\x0fpw.thread.protoB\x06Threadb\x06proto3'
  ,
  dependencies=[pw__thread__protos_dot_thread__pb2.DESCRIPTOR,])




_THREADREQUEST = _descriptor.Descriptor(
  name='ThreadRequest',
  full_name='pw.thread.proto.ThreadRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='pw.thread.proto.ThreadRequest.name', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_name', full_name='pw.thread.proto.ThreadRequest._name',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=98,
  serialized_end=141,
)

_THREADREQUEST.oneofs_by_name['_name'].fields.append(
  _THREADREQUEST.fields_by_name['name'])
_THREADREQUEST.fields_by_name['name'].containing_oneof = _THREADREQUEST.oneofs_by_name['_name']
DESCRIPTOR.message_types_by_name['ThreadRequest'] = _THREADREQUEST
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ThreadRequest = _reflection.GeneratedProtocolMessageType('ThreadRequest', (_message.Message,), {
  'DESCRIPTOR' : _THREADREQUEST,
  '__module__' : 'pw_thread_protos.thread_snapshot_service_pb2'
  # @@protoc_insertion_point(class_scope:pw.thread.proto.ThreadRequest)
  })
_sym_db.RegisterMessage(ThreadRequest)


DESCRIPTOR._options = None

_THREADSNAPSHOTSERVICE = _descriptor.ServiceDescriptor(
  name='ThreadSnapshotService',
  full_name='pw.thread.proto.ThreadSnapshotService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=143,
  serialized_end=260,
  methods=[
  _descriptor.MethodDescriptor(
    name='GetPeakStackUsage',
    full_name='pw.thread.proto.ThreadSnapshotService.GetPeakStackUsage',
    index=0,
    containing_service=None,
    input_type=_THREADREQUEST,
    output_type=pw__thread__protos_dot_thread__pb2._SNAPSHOTTHREADINFO,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_THREADSNAPSHOTSERVICE)

DESCRIPTOR.services_by_name['ThreadSnapshotService'] = _THREADSNAPSHOTSERVICE

# @@protoc_insertion_point(module_scope)
