# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_tokenizer/proto/options.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_tokenizer/proto/options.proto',
  package='pw.tokenizer',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n pw_tokenizer/proto/options.proto\x12\x0cpw.tokenizer\x1a google/protobuf/descriptor.proto*)\n\x0cTokenization\x12\x19\n\x15TOKENIZATION_OPTIONAL\x10\x00:K\n\x06\x66ormat\x12\x1d.google.protobuf.FieldOptions\x18\xf0\xe5\x04 \x01(\x0e\x32\x1a.pw.tokenizer.Tokenizationb\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_descriptor__pb2.DESCRIPTOR,])

_TOKENIZATION = _descriptor.EnumDescriptor(
  name='Tokenization',
  full_name='pw.tokenizer.Tokenization',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='TOKENIZATION_OPTIONAL', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=84,
  serialized_end=125,
)
_sym_db.RegisterEnumDescriptor(_TOKENIZATION)

Tokenization = enum_type_wrapper.EnumTypeWrapper(_TOKENIZATION)
TOKENIZATION_OPTIONAL = 0

FORMAT_FIELD_NUMBER = 78576
format = _descriptor.FieldDescriptor(
  name='format', full_name='pw.tokenizer.format', index=0,
  number=78576, type=14, cpp_type=8, label=1,
  has_default_value=False, default_value=0,
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key)

DESCRIPTOR.enum_types_by_name['Tokenization'] = _TOKENIZATION
DESCRIPTOR.extensions_by_name['format'] = format
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

format.enum_type = _TOKENIZATION
google_dot_protobuf_dot_descriptor__pb2.FieldOptions.RegisterExtension(format)

# @@protoc_insertion_point(module_scope)
