# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_snapshot_metadata_proto/snapshot_metadata.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pw_tokenizer.proto import options_pb2 as pw__tokenizer_dot_proto_dot_options__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_snapshot_metadata_proto/snapshot_metadata.proto',
  package='pw.snapshot',
  syntax='proto3',
  serialized_options=b'\n\021pw.snapshot.protoB\010Snapshot',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n2pw_snapshot_metadata_proto/snapshot_metadata.proto\x12\x0bpw.snapshot\x1a pw_tokenizer/proto/options.proto\"J\n\x0f\x43puArchitecture\"7\n\x04\x45num\x12\x0b\n\x07UNKNOWN\x10\x00\x12\n\n\x06\x41RMV6M\x10\x01\x12\n\n\x06\x41RMV7M\x10\x02\x12\n\n\x06\x41RMV8M\x10\x03\"\xe9\x01\n\x08Metadata\x12\x14\n\x06reason\x18\x01 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\r\n\x05\x66\x61tal\x18\x02 \x01(\x08\x12\x1a\n\x0cproject_name\x18\x03 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\x18\n\x10software_version\x18\x04 \x01(\t\x12\x1b\n\x13software_build_uuid\x18\x05 \x01(\x0c\x12\x19\n\x0b\x64\x65vice_name\x18\x06 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\x15\n\rsnapshot_uuid\x18\x07 \x01(\x0c\x12\x33\n\x08\x63pu_arch\x18\x08 \x01(\x0e\x32!.pw.snapshot.CpuArchitecture.Enum\"\xa1\x01\n\x11SnapshotBasicInfo\x12\'\n\x08metadata\x18\x10 \x01(\x0b\x32\x15.pw.snapshot.Metadata\x12\x36\n\x04tags\x18\x11 \x03(\x0b\x32(.pw.snapshot.SnapshotBasicInfo.TagsEntry\x1a+\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\x1d\n\x11pw.snapshot.protoB\x08Snapshotb\x06proto3'
  ,
  dependencies=[pw__tokenizer_dot_proto_dot_options__pb2.DESCRIPTOR,])



_CPUARCHITECTURE_ENUM = _descriptor.EnumDescriptor(
  name='Enum',
  full_name='pw.snapshot.CpuArchitecture.Enum',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ARMV6M', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ARMV7M', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ARMV8M', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=120,
  serialized_end=175,
)
_sym_db.RegisterEnumDescriptor(_CPUARCHITECTURE_ENUM)


_CPUARCHITECTURE = _descriptor.Descriptor(
  name='CpuArchitecture',
  full_name='pw.snapshot.CpuArchitecture',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _CPUARCHITECTURE_ENUM,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=101,
  serialized_end=175,
)


_METADATA = _descriptor.Descriptor(
  name='Metadata',
  full_name='pw.snapshot.Metadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='reason', full_name='pw.snapshot.Metadata.reason', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\200\257&\000', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fatal', full_name='pw.snapshot.Metadata.fatal', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='project_name', full_name='pw.snapshot.Metadata.project_name', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\200\257&\000', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='software_version', full_name='pw.snapshot.Metadata.software_version', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='software_build_uuid', full_name='pw.snapshot.Metadata.software_build_uuid', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='device_name', full_name='pw.snapshot.Metadata.device_name', index=5,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\200\257&\000', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='snapshot_uuid', full_name='pw.snapshot.Metadata.snapshot_uuid', index=6,
      number=7, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cpu_arch', full_name='pw.snapshot.Metadata.cpu_arch', index=7,
      number=8, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=178,
  serialized_end=411,
)


_SNAPSHOTBASICINFO_TAGSENTRY = _descriptor.Descriptor(
  name='TagsEntry',
  full_name='pw.snapshot.SnapshotBasicInfo.TagsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='pw.snapshot.SnapshotBasicInfo.TagsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='pw.snapshot.SnapshotBasicInfo.TagsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=532,
  serialized_end=575,
)

_SNAPSHOTBASICINFO = _descriptor.Descriptor(
  name='SnapshotBasicInfo',
  full_name='pw.snapshot.SnapshotBasicInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metadata', full_name='pw.snapshot.SnapshotBasicInfo.metadata', index=0,
      number=16, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tags', full_name='pw.snapshot.SnapshotBasicInfo.tags', index=1,
      number=17, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_SNAPSHOTBASICINFO_TAGSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=414,
  serialized_end=575,
)

_CPUARCHITECTURE_ENUM.containing_type = _CPUARCHITECTURE
_METADATA.fields_by_name['cpu_arch'].enum_type = _CPUARCHITECTURE_ENUM
_SNAPSHOTBASICINFO_TAGSENTRY.containing_type = _SNAPSHOTBASICINFO
_SNAPSHOTBASICINFO.fields_by_name['metadata'].message_type = _METADATA
_SNAPSHOTBASICINFO.fields_by_name['tags'].message_type = _SNAPSHOTBASICINFO_TAGSENTRY
DESCRIPTOR.message_types_by_name['CpuArchitecture'] = _CPUARCHITECTURE
DESCRIPTOR.message_types_by_name['Metadata'] = _METADATA
DESCRIPTOR.message_types_by_name['SnapshotBasicInfo'] = _SNAPSHOTBASICINFO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

CpuArchitecture = _reflection.GeneratedProtocolMessageType('CpuArchitecture', (_message.Message,), {
  'DESCRIPTOR' : _CPUARCHITECTURE,
  '__module__' : 'pw_snapshot_metadata_proto.snapshot_metadata_pb2'
  # @@protoc_insertion_point(class_scope:pw.snapshot.CpuArchitecture)
  })
_sym_db.RegisterMessage(CpuArchitecture)

Metadata = _reflection.GeneratedProtocolMessageType('Metadata', (_message.Message,), {
  'DESCRIPTOR' : _METADATA,
  '__module__' : 'pw_snapshot_metadata_proto.snapshot_metadata_pb2'
  # @@protoc_insertion_point(class_scope:pw.snapshot.Metadata)
  })
_sym_db.RegisterMessage(Metadata)

SnapshotBasicInfo = _reflection.GeneratedProtocolMessageType('SnapshotBasicInfo', (_message.Message,), {

  'TagsEntry' : _reflection.GeneratedProtocolMessageType('TagsEntry', (_message.Message,), {
    'DESCRIPTOR' : _SNAPSHOTBASICINFO_TAGSENTRY,
    '__module__' : 'pw_snapshot_metadata_proto.snapshot_metadata_pb2'
    # @@protoc_insertion_point(class_scope:pw.snapshot.SnapshotBasicInfo.TagsEntry)
    })
  ,
  'DESCRIPTOR' : _SNAPSHOTBASICINFO,
  '__module__' : 'pw_snapshot_metadata_proto.snapshot_metadata_pb2'
  # @@protoc_insertion_point(class_scope:pw.snapshot.SnapshotBasicInfo)
  })
_sym_db.RegisterMessage(SnapshotBasicInfo)
_sym_db.RegisterMessage(SnapshotBasicInfo.TagsEntry)


DESCRIPTOR._options = None
_METADATA.fields_by_name['reason']._options = None
_METADATA.fields_by_name['project_name']._options = None
_METADATA.fields_by_name['device_name']._options = None
_SNAPSHOTBASICINFO_TAGSENTRY._options = None
# @@protoc_insertion_point(module_scope)
