# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_chrono_protos/chrono.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_chrono_protos/chrono.proto',
  package='pw.chrono',
  syntax='proto3',
  serialized_options=b'B\006Chrono',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1dpw_chrono_protos/chrono.proto\x12\tpw.chrono\"q\n\tEpochType\"d\n\x04\x45num\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x13\n\x0fTIME_SINCE_BOOT\x10\x01\x12\x12\n\x0eUTC_WALL_CLOCK\x10\x02\x12\x12\n\x0eGPS_WALL_CLOCK\x10\x03\x12\x12\n\x0eTAI_WALL_CLOCK\x10\x04\"\xa4\x01\n\x0f\x43lockParameters\x12%\n\x1dtick_period_seconds_numerator\x18\x01 \x01(\x05\x12\'\n\x1ftick_period_seconds_denominator\x18\x02 \x01(\x05\x12\x32\n\nepoch_type\x18\x03 \x01(\x0e\x32\x19.pw.chrono.EpochType.EnumH\x00\x88\x01\x01\x42\r\n\x0b_epoch_type\"T\n\tTimePoint\x12\x11\n\ttimestamp\x18\x01 \x01(\x03\x12\x34\n\x10\x63lock_parameters\x18\x02 \x01(\x0b\x32\x1a.pw.chrono.ClockParameters\">\n\x12SnapshotTimestamps\x12(\n\ntimestamps\x18\x16 \x03(\x0b\x32\x14.pw.chrono.TimePointB\x08\x42\x06\x43hronob\x06proto3'
)



_EPOCHTYPE_ENUM = _descriptor.EnumDescriptor(
  name='Enum',
  full_name='pw.chrono.EpochType.Enum',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TIME_SINCE_BOOT', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UTC_WALL_CLOCK', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='GPS_WALL_CLOCK', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TAI_WALL_CLOCK', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=57,
  serialized_end=157,
)
_sym_db.RegisterEnumDescriptor(_EPOCHTYPE_ENUM)


_EPOCHTYPE = _descriptor.Descriptor(
  name='EpochType',
  full_name='pw.chrono.EpochType',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _EPOCHTYPE_ENUM,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=44,
  serialized_end=157,
)


_CLOCKPARAMETERS = _descriptor.Descriptor(
  name='ClockParameters',
  full_name='pw.chrono.ClockParameters',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='tick_period_seconds_numerator', full_name='pw.chrono.ClockParameters.tick_period_seconds_numerator', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tick_period_seconds_denominator', full_name='pw.chrono.ClockParameters.tick_period_seconds_denominator', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='epoch_type', full_name='pw.chrono.ClockParameters.epoch_type', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_epoch_type', full_name='pw.chrono.ClockParameters._epoch_type',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=160,
  serialized_end=324,
)


_TIMEPOINT = _descriptor.Descriptor(
  name='TimePoint',
  full_name='pw.chrono.TimePoint',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='pw.chrono.TimePoint.timestamp', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='clock_parameters', full_name='pw.chrono.TimePoint.clock_parameters', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=326,
  serialized_end=410,
)


_SNAPSHOTTIMESTAMPS = _descriptor.Descriptor(
  name='SnapshotTimestamps',
  full_name='pw.chrono.SnapshotTimestamps',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='timestamps', full_name='pw.chrono.SnapshotTimestamps.timestamps', index=0,
      number=22, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=412,
  serialized_end=474,
)

_EPOCHTYPE_ENUM.containing_type = _EPOCHTYPE
_CLOCKPARAMETERS.fields_by_name['epoch_type'].enum_type = _EPOCHTYPE_ENUM
_CLOCKPARAMETERS.oneofs_by_name['_epoch_type'].fields.append(
  _CLOCKPARAMETERS.fields_by_name['epoch_type'])
_CLOCKPARAMETERS.fields_by_name['epoch_type'].containing_oneof = _CLOCKPARAMETERS.oneofs_by_name['_epoch_type']
_TIMEPOINT.fields_by_name['clock_parameters'].message_type = _CLOCKPARAMETERS
_SNAPSHOTTIMESTAMPS.fields_by_name['timestamps'].message_type = _TIMEPOINT
DESCRIPTOR.message_types_by_name['EpochType'] = _EPOCHTYPE
DESCRIPTOR.message_types_by_name['ClockParameters'] = _CLOCKPARAMETERS
DESCRIPTOR.message_types_by_name['TimePoint'] = _TIMEPOINT
DESCRIPTOR.message_types_by_name['SnapshotTimestamps'] = _SNAPSHOTTIMESTAMPS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

EpochType = _reflection.GeneratedProtocolMessageType('EpochType', (_message.Message,), {
  'DESCRIPTOR' : _EPOCHTYPE,
  '__module__' : 'pw_chrono_protos.chrono_pb2'
  # @@protoc_insertion_point(class_scope:pw.chrono.EpochType)
  })
_sym_db.RegisterMessage(EpochType)

ClockParameters = _reflection.GeneratedProtocolMessageType('ClockParameters', (_message.Message,), {
  'DESCRIPTOR' : _CLOCKPARAMETERS,
  '__module__' : 'pw_chrono_protos.chrono_pb2'
  # @@protoc_insertion_point(class_scope:pw.chrono.ClockParameters)
  })
_sym_db.RegisterMessage(ClockParameters)

TimePoint = _reflection.GeneratedProtocolMessageType('TimePoint', (_message.Message,), {
  'DESCRIPTOR' : _TIMEPOINT,
  '__module__' : 'pw_chrono_protos.chrono_pb2'
  # @@protoc_insertion_point(class_scope:pw.chrono.TimePoint)
  })
_sym_db.RegisterMessage(TimePoint)

SnapshotTimestamps = _reflection.GeneratedProtocolMessageType('SnapshotTimestamps', (_message.Message,), {
  'DESCRIPTOR' : _SNAPSHOTTIMESTAMPS,
  '__module__' : 'pw_chrono_protos.chrono_pb2'
  # @@protoc_insertion_point(class_scope:pw.chrono.SnapshotTimestamps)
  })
_sym_db.RegisterMessage(SnapshotTimestamps)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)
