# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_protobuf_codegen_protos/codegen_options.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_protobuf_codegen_protos/codegen_options.proto',
  package='pw.protobuf',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n0pw_protobuf_codegen_protos/codegen_options.proto\x12\x0bpw.protobuf\"t\n\x0e\x43odegenOptions\x12\x11\n\tmax_count\x18\x01 \x01(\x05\x12\x13\n\x0b\x66ixed_count\x18\x02 \x01(\x08\x12\x10\n\x08max_size\x18\x03 \x01(\x05\x12\x12\n\nfixed_size\x18\x04 \x01(\x08\x12\x14\n\x0cuse_callback\x18\x05 \x01(\x08\x62\x06proto3'
)




_CODEGENOPTIONS = _descriptor.Descriptor(
  name='CodegenOptions',
  full_name='pw.protobuf.CodegenOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='max_count', full_name='pw.protobuf.CodegenOptions.max_count', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fixed_count', full_name='pw.protobuf.CodegenOptions.fixed_count', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max_size', full_name='pw.protobuf.CodegenOptions.max_size', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fixed_size', full_name='pw.protobuf.CodegenOptions.fixed_size', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='use_callback', full_name='pw.protobuf.CodegenOptions.use_callback', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=65,
  serialized_end=181,
)

DESCRIPTOR.message_types_by_name['CodegenOptions'] = _CODEGENOPTIONS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

CodegenOptions = _reflection.GeneratedProtocolMessageType('CodegenOptions', (_message.Message,), {
  'DESCRIPTOR' : _CODEGENOPTIONS,
  '__module__' : 'pw_protobuf_codegen_protos.codegen_options_pb2'
  # @@protoc_insertion_point(class_scope:pw.protobuf.CodegenOptions)
  })
_sym_db.RegisterMessage(CodegenOptions)


# @@protoc_insertion_point(module_scope)
