# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_log/proto/log.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pw_protobuf_protos import common_pb2 as pw__protobuf__protos_dot_common__pb2
from pw_tokenizer.proto import options_pb2 as pw__tokenizer_dot_proto_dot_options__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_log/proto/log.proto',
  package='pw.log',
  syntax='proto3',
  serialized_options=b'B\003Log',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x16pw_log/proto/log.proto\x12\x06pw.log\x1a\x1fpw_protobuf_protos/common.proto\x1a pw_tokenizer/proto/options.proto\"\xe4\x01\n\x08LogEntry\x12\x1a\n\x07message\x18\x01 \x01(\x0c\x42\x04\x80\xaf&\x00H\x01\x88\x01\x01\x12\x12\n\nline_level\x18\x02 \x01(\r\x12\r\n\x05\x66lags\x18\x03 \x01(\r\x12\x13\n\ttimestamp\x18\x04 \x01(\x03H\x00\x12\x1f\n\x15time_since_last_entry\x18\x05 \x01(\x03H\x00\x12\x0f\n\x07\x64ropped\x18\x06 \x01(\r\x12\x14\n\x06module\x18\x07 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\x12\n\x04\x66ile\x18\x08 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\x14\n\x06thread\x18\t \x01(\x0c\x42\x04\x80\xaf&\x00\x42\x06\n\x04timeB\n\n\x08_message\"\x0c\n\nLogRequest\"P\n\nLogEntries\x12!\n\x07\x65ntries\x18\x01 \x03(\x0b\x32\x10.pw.log.LogEntry\x12\x1f\n\x17\x66irst_entry_sequence_id\x18\x02 \x01(\r\"\xf2\x02\n\nFilterRule\x12=\n\x1blevel_greater_than_or_equal\x18\x01 \x01(\x0e\x32\x18.pw.log.FilterRule.Level\x12\x1b\n\rmodule_equals\x18\x02 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\x15\n\rany_flags_set\x18\x03 \x01(\r\x12)\n\x06\x61\x63tion\x18\x04 \x01(\x0e\x32\x19.pw.log.FilterRule.Action\x12\x1b\n\rthread_equals\x18\x05 \x01(\x0c\x42\x04\x80\xaf&\x00\"}\n\x05Level\x12\r\n\tANY_LEVEL\x10\x00\x12\x0f\n\x0b\x44\x45\x42UG_LEVEL\x10\x01\x12\x0e\n\nINFO_LEVEL\x10\x02\x12\x0e\n\nWARN_LEVEL\x10\x03\x12\x0f\n\x0b\x45RROR_LEVEL\x10\x04\x12\x12\n\x0e\x43RITICAL_LEVEL\x10\x05\x12\x0f\n\x0b\x46\x41TAL_LEVEL\x10\x07\"*\n\x06\x41\x63tion\x12\x0c\n\x08INACTIVE\x10\x00\x12\x08\n\x04KEEP\x10\x01\x12\x08\n\x04\x44ROP\x10\x02\"*\n\x06\x46ilter\x12 \n\x04rule\x18\x01 \x03(\x0b\x32\x12.pw.log.FilterRule\"K\n\x10SetFilterRequest\x12\x17\n\tfilter_id\x18\x01 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\x1e\n\x06\x66ilter\x18\x02 \x01(\x0b\x32\x0e.pw.log.Filter\"+\n\x10GetFilterRequest\x12\x17\n\tfilter_id\x18\x01 \x01(\x0c\x42\x04\x80\xaf&\x00\"\x15\n\x13\x46ilterIdListRequest\"/\n\x14\x46ilterIdListResponse\x12\x17\n\tfilter_id\x18\x01 \x03(\x0c\x42\x04\x80\xaf&\x00\x32:\n\x04Logs\x12\x32\n\x06Listen\x12\x12.pw.log.LogRequest\x1a\x12.pw.log.LogEntries0\x01\x32\xc7\x01\n\x07\x46ilters\x12\x39\n\tSetFilter\x12\x18.pw.log.SetFilterRequest\x1a\x12.pw.protobuf.Empty\x12\x35\n\tGetFilter\x12\x18.pw.log.GetFilterRequest\x1a\x0e.pw.log.Filter\x12J\n\rListFilterIds\x12\x1b.pw.log.FilterIdListRequest\x1a\x1c.pw.log.FilterIdListResponseB\x05\x42\x03Logb\x06proto3'
  ,
  dependencies=[pw__protobuf__protos_dot_common__pb2.DESCRIPTOR,pw__tokenizer_dot_proto_dot_options__pb2.DESCRIPTOR,])



_FILTERRULE_LEVEL = _descriptor.EnumDescriptor(
  name='Level',
  full_name='pw.log.FilterRule.Level',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='ANY_LEVEL', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DEBUG_LEVEL', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INFO_LEVEL', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='WARN_LEVEL', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ERROR_LEVEL', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CRITICAL_LEVEL', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FATAL_LEVEL', index=6, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=630,
  serialized_end=755,
)
_sym_db.RegisterEnumDescriptor(_FILTERRULE_LEVEL)

_FILTERRULE_ACTION = _descriptor.EnumDescriptor(
  name='Action',
  full_name='pw.log.FilterRule.Action',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='INACTIVE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='KEEP', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DROP', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=757,
  serialized_end=799,
)
_sym_db.RegisterEnumDescriptor(_FILTERRULE_ACTION)


_LOGENTRY = _descriptor.Descriptor(
  name='LogEntry',
  full_name='pw.log.LogEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='message', full_name='pw.log.LogEntry.message', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\200\257&\000', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='line_level', full_name='pw.log.LogEntry.line_level', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='flags', full_name='pw.log.LogEntry.flags', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='pw.log.LogEntry.timestamp', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='time_since_last_entry', full_name='pw.log.LogEntry.time_since_last_entry', index=4,
      number=5, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dropped', full_name='pw.log.LogEntry.dropped', index=5,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='module', full_name='pw.log.LogEntry.module', index=6,
      number=7, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\200\257&\000', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='file', full_name='pw.log.LogEntry.file', index=7,
      number=8, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\200\257&\000', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='thread', full_name='pw.log.LogEntry.thread', index=8,
      number=9, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\200\257&\000', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='time', full_name='pw.log.LogEntry.time',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_message', full_name='pw.log.LogEntry._message',
      index=1, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=102,
  serialized_end=330,
)


_LOGREQUEST = _descriptor.Descriptor(
  name='LogRequest',
  full_name='pw.log.LogRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=332,
  serialized_end=344,
)


_LOGENTRIES = _descriptor.Descriptor(
  name='LogEntries',
  full_name='pw.log.LogEntries',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='entries', full_name='pw.log.LogEntries.entries', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='first_entry_sequence_id', full_name='pw.log.LogEntries.first_entry_sequence_id', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=346,
  serialized_end=426,
)


_FILTERRULE = _descriptor.Descriptor(
  name='FilterRule',
  full_name='pw.log.FilterRule',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='level_greater_than_or_equal', full_name='pw.log.FilterRule.level_greater_than_or_equal', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='module_equals', full_name='pw.log.FilterRule.module_equals', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\200\257&\000', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='any_flags_set', full_name='pw.log.FilterRule.any_flags_set', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='action', full_name='pw.log.FilterRule.action', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='thread_equals', full_name='pw.log.FilterRule.thread_equals', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\200\257&\000', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _FILTERRULE_LEVEL,
    _FILTERRULE_ACTION,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=429,
  serialized_end=799,
)


_FILTER = _descriptor.Descriptor(
  name='Filter',
  full_name='pw.log.Filter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='rule', full_name='pw.log.Filter.rule', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=801,
  serialized_end=843,
)


_SETFILTERREQUEST = _descriptor.Descriptor(
  name='SetFilterRequest',
  full_name='pw.log.SetFilterRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='filter_id', full_name='pw.log.SetFilterRequest.filter_id', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\200\257&\000', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filter', full_name='pw.log.SetFilterRequest.filter', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=845,
  serialized_end=920,
)


_GETFILTERREQUEST = _descriptor.Descriptor(
  name='GetFilterRequest',
  full_name='pw.log.GetFilterRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='filter_id', full_name='pw.log.GetFilterRequest.filter_id', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\200\257&\000', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=922,
  serialized_end=965,
)


_FILTERIDLISTREQUEST = _descriptor.Descriptor(
  name='FilterIdListRequest',
  full_name='pw.log.FilterIdListRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=967,
  serialized_end=988,
)


_FILTERIDLISTRESPONSE = _descriptor.Descriptor(
  name='FilterIdListResponse',
  full_name='pw.log.FilterIdListResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='filter_id', full_name='pw.log.FilterIdListResponse.filter_id', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\200\257&\000', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=990,
  serialized_end=1037,
)

_LOGENTRY.oneofs_by_name['time'].fields.append(
  _LOGENTRY.fields_by_name['timestamp'])
_LOGENTRY.fields_by_name['timestamp'].containing_oneof = _LOGENTRY.oneofs_by_name['time']
_LOGENTRY.oneofs_by_name['time'].fields.append(
  _LOGENTRY.fields_by_name['time_since_last_entry'])
_LOGENTRY.fields_by_name['time_since_last_entry'].containing_oneof = _LOGENTRY.oneofs_by_name['time']
_LOGENTRY.oneofs_by_name['_message'].fields.append(
  _LOGENTRY.fields_by_name['message'])
_LOGENTRY.fields_by_name['message'].containing_oneof = _LOGENTRY.oneofs_by_name['_message']
_LOGENTRIES.fields_by_name['entries'].message_type = _LOGENTRY
_FILTERRULE.fields_by_name['level_greater_than_or_equal'].enum_type = _FILTERRULE_LEVEL
_FILTERRULE.fields_by_name['action'].enum_type = _FILTERRULE_ACTION
_FILTERRULE_LEVEL.containing_type = _FILTERRULE
_FILTERRULE_ACTION.containing_type = _FILTERRULE
_FILTER.fields_by_name['rule'].message_type = _FILTERRULE
_SETFILTERREQUEST.fields_by_name['filter'].message_type = _FILTER
DESCRIPTOR.message_types_by_name['LogEntry'] = _LOGENTRY
DESCRIPTOR.message_types_by_name['LogRequest'] = _LOGREQUEST
DESCRIPTOR.message_types_by_name['LogEntries'] = _LOGENTRIES
DESCRIPTOR.message_types_by_name['FilterRule'] = _FILTERRULE
DESCRIPTOR.message_types_by_name['Filter'] = _FILTER
DESCRIPTOR.message_types_by_name['SetFilterRequest'] = _SETFILTERREQUEST
DESCRIPTOR.message_types_by_name['GetFilterRequest'] = _GETFILTERREQUEST
DESCRIPTOR.message_types_by_name['FilterIdListRequest'] = _FILTERIDLISTREQUEST
DESCRIPTOR.message_types_by_name['FilterIdListResponse'] = _FILTERIDLISTRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

LogEntry = _reflection.GeneratedProtocolMessageType('LogEntry', (_message.Message,), {
  'DESCRIPTOR' : _LOGENTRY,
  '__module__' : 'pw_log.proto.log_pb2'
  # @@protoc_insertion_point(class_scope:pw.log.LogEntry)
  })
_sym_db.RegisterMessage(LogEntry)

LogRequest = _reflection.GeneratedProtocolMessageType('LogRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOGREQUEST,
  '__module__' : 'pw_log.proto.log_pb2'
  # @@protoc_insertion_point(class_scope:pw.log.LogRequest)
  })
_sym_db.RegisterMessage(LogRequest)

LogEntries = _reflection.GeneratedProtocolMessageType('LogEntries', (_message.Message,), {
  'DESCRIPTOR' : _LOGENTRIES,
  '__module__' : 'pw_log.proto.log_pb2'
  # @@protoc_insertion_point(class_scope:pw.log.LogEntries)
  })
_sym_db.RegisterMessage(LogEntries)

FilterRule = _reflection.GeneratedProtocolMessageType('FilterRule', (_message.Message,), {
  'DESCRIPTOR' : _FILTERRULE,
  '__module__' : 'pw_log.proto.log_pb2'
  # @@protoc_insertion_point(class_scope:pw.log.FilterRule)
  })
_sym_db.RegisterMessage(FilterRule)

Filter = _reflection.GeneratedProtocolMessageType('Filter', (_message.Message,), {
  'DESCRIPTOR' : _FILTER,
  '__module__' : 'pw_log.proto.log_pb2'
  # @@protoc_insertion_point(class_scope:pw.log.Filter)
  })
_sym_db.RegisterMessage(Filter)

SetFilterRequest = _reflection.GeneratedProtocolMessageType('SetFilterRequest', (_message.Message,), {
  'DESCRIPTOR' : _SETFILTERREQUEST,
  '__module__' : 'pw_log.proto.log_pb2'
  # @@protoc_insertion_point(class_scope:pw.log.SetFilterRequest)
  })
_sym_db.RegisterMessage(SetFilterRequest)

GetFilterRequest = _reflection.GeneratedProtocolMessageType('GetFilterRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETFILTERREQUEST,
  '__module__' : 'pw_log.proto.log_pb2'
  # @@protoc_insertion_point(class_scope:pw.log.GetFilterRequest)
  })
_sym_db.RegisterMessage(GetFilterRequest)

FilterIdListRequest = _reflection.GeneratedProtocolMessageType('FilterIdListRequest', (_message.Message,), {
  'DESCRIPTOR' : _FILTERIDLISTREQUEST,
  '__module__' : 'pw_log.proto.log_pb2'
  # @@protoc_insertion_point(class_scope:pw.log.FilterIdListRequest)
  })
_sym_db.RegisterMessage(FilterIdListRequest)

FilterIdListResponse = _reflection.GeneratedProtocolMessageType('FilterIdListResponse', (_message.Message,), {
  'DESCRIPTOR' : _FILTERIDLISTRESPONSE,
  '__module__' : 'pw_log.proto.log_pb2'
  # @@protoc_insertion_point(class_scope:pw.log.FilterIdListResponse)
  })
_sym_db.RegisterMessage(FilterIdListResponse)


DESCRIPTOR._options = None
_LOGENTRY.fields_by_name['message']._options = None
_LOGENTRY.fields_by_name['module']._options = None
_LOGENTRY.fields_by_name['file']._options = None
_LOGENTRY.fields_by_name['thread']._options = None
_FILTERRULE.fields_by_name['module_equals']._options = None
_FILTERRULE.fields_by_name['thread_equals']._options = None
_SETFILTERREQUEST.fields_by_name['filter_id']._options = None
_GETFILTERREQUEST.fields_by_name['filter_id']._options = None
_FILTERIDLISTRESPONSE.fields_by_name['filter_id']._options = None

_LOGS = _descriptor.ServiceDescriptor(
  name='Logs',
  full_name='pw.log.Logs',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=1039,
  serialized_end=1097,
  methods=[
  _descriptor.MethodDescriptor(
    name='Listen',
    full_name='pw.log.Logs.Listen',
    index=0,
    containing_service=None,
    input_type=_LOGREQUEST,
    output_type=_LOGENTRIES,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_LOGS)

DESCRIPTOR.services_by_name['Logs'] = _LOGS


_FILTERS = _descriptor.ServiceDescriptor(
  name='Filters',
  full_name='pw.log.Filters',
  file=DESCRIPTOR,
  index=1,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=1100,
  serialized_end=1299,
  methods=[
  _descriptor.MethodDescriptor(
    name='SetFilter',
    full_name='pw.log.Filters.SetFilter',
    index=0,
    containing_service=None,
    input_type=_SETFILTERREQUEST,
    output_type=pw__protobuf__protos_dot_common__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetFilter',
    full_name='pw.log.Filters.GetFilter',
    index=1,
    containing_service=None,
    input_type=_GETFILTERREQUEST,
    output_type=_FILTER,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListFilterIds',
    full_name='pw.log.Filters.ListFilterIds',
    index=2,
    containing_service=None,
    input_type=_FILTERIDLISTREQUEST,
    output_type=_FILTERIDLISTRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_FILTERS)

DESCRIPTOR.services_by_name['Filters'] = _FILTERS

# @@protoc_insertion_point(module_scope)
