# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_protobuf_protos/field_options.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_protobuf_protos/field_options.proto',
  package='pw.protobuf',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n&pw_protobuf_protos/field_options.proto\x12\x0bpw.protobuf\x1a google/protobuf/descriptor.proto\"X\n\x0c\x46ieldOptions\x12\x16\n\tmax_count\x18\x01 \x01(\x05H\x00\x88\x01\x01\x12\x15\n\x08max_size\x18\x02 \x01(\x05H\x01\x88\x01\x01\x42\x0c\n\n_max_countB\x0b\n\t_max_size:J\n\x04pwpb\x12\x1d.google.protobuf.FieldOptions\x18\x83\t \x01(\x0b\x32\x19.pw.protobuf.FieldOptions\x88\x01\x01\x62\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_descriptor__pb2.DESCRIPTOR,])


PWPB_FIELD_NUMBER = 1155
pwpb = _descriptor.FieldDescriptor(
  name='pwpb', full_name='pw.protobuf.pwpb', index=0,
  number=1155, type=11, cpp_type=10, label=1,
  has_default_value=False, default_value=None,
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key)


_FIELDOPTIONS = _descriptor.Descriptor(
  name='FieldOptions',
  full_name='pw.protobuf.FieldOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='max_count', full_name='pw.protobuf.FieldOptions.max_count', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max_size', full_name='pw.protobuf.FieldOptions.max_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_max_count', full_name='pw.protobuf.FieldOptions._max_count',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_max_size', full_name='pw.protobuf.FieldOptions._max_size',
      index=1, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=89,
  serialized_end=177,
)

_FIELDOPTIONS.oneofs_by_name['_max_count'].fields.append(
  _FIELDOPTIONS.fields_by_name['max_count'])
_FIELDOPTIONS.fields_by_name['max_count'].containing_oneof = _FIELDOPTIONS.oneofs_by_name['_max_count']
_FIELDOPTIONS.oneofs_by_name['_max_size'].fields.append(
  _FIELDOPTIONS.fields_by_name['max_size'])
_FIELDOPTIONS.fields_by_name['max_size'].containing_oneof = _FIELDOPTIONS.oneofs_by_name['_max_size']
DESCRIPTOR.message_types_by_name['FieldOptions'] = _FIELDOPTIONS
DESCRIPTOR.extensions_by_name['pwpb'] = pwpb
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

FieldOptions = _reflection.GeneratedProtocolMessageType('FieldOptions', (_message.Message,), {
  'DESCRIPTOR' : _FIELDOPTIONS,
  '__module__' : 'pw_protobuf_protos.field_options_pb2'
  # @@protoc_insertion_point(class_scope:pw.protobuf.FieldOptions)
  })
_sym_db.RegisterMessage(FieldOptions)

pwpb.message_type = _FIELDOPTIONS
google_dot_protobuf_dot_descriptor__pb2.FieldOptions.RegisterExtension(pwpb)

# @@protoc_insertion_point(module_scope)
