# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_software_update/bundled_update.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pw_protobuf_protos import common_pb2 as pw__protobuf__protos_dot_common__pb2
from pw_tokenizer.proto import options_pb2 as pw__tokenizer_dot_proto_dot_options__pb2
from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_software_update/bundled_update.proto',
  package='pw.software_update',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\'pw_software_update/bundled_update.proto\x12\x12pw.software_update\x1a\x1fpw_protobuf_protos/common.proto\x1a pw_tokenizer/proto/options.proto\x1a\x19google/protobuf/any.proto\"\xa2\x01\n\x12\x42undledUpdateState\"\x8b\x01\n\x04\x45num\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0c\n\x08INACTIVE\x10\x01\x12\x10\n\x0cTRANSFERRING\x10\x02\x12\x0f\n\x0bTRANSFERRED\x10\x03\x12\r\n\tVERIFYING\x10\x04\x12\x0c\n\x08VERIFIED\x10\x05\x12\x0c\n\x08\x41PPLYING\x10\x06\x12\x0c\n\x08\x41\x42ORTING\x10\x07\x12\x0c\n\x08\x46INISHED\x10\x08\"\x91\x01\n\x13\x42undledUpdateResult\"z\n\x04\x45num\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0b\n\x07SUCCESS\x10\x01\x12\x11\n\rUNKNOWN_ERROR\x10\x02\x12\x0b\n\x07\x41\x42ORTED\x10\x03\x12\x13\n\x0fTRANSFER_FAILED\x10\x04\x12\x11\n\rVERIFY_FAILED\x10\x05\x12\x10\n\x0c\x41PPLY_FAILED\x10\x06\"\xc7\x03\n\x13\x42undledUpdateStatus\x12:\n\x05state\x18\x01 \x01(\x0e\x32+.pw.software_update.BundledUpdateState.Enum\x12\x41\n\x06result\x18\x02 \x01(\x0e\x32,.pw.software_update.BundledUpdateResult.EnumH\x00\x88\x01\x01\x12\x34\n\'current_state_progress_hundreth_percent\x18\x03 \x01(\rH\x01\x88\x01\x01\x12\x18\n\x0btransfer_id\x18\x04 \x01(\rH\x02\x88\x01\x01\x12\x1c\n\x0f\x62undle_filename\x18\x05 \x01(\tH\x03\x88\x01\x01\x12\x17\n\x04note\x18\x06 \x01(\x0c\x42\x04\x80\xaf&\x00H\x04\x88\x01\x01\x12\x32\n\x0f\x65xtended_status\x18\x07 \x01(\x0b\x32\x14.google.protobuf.AnyH\x05\x88\x01\x01\x42\t\n\x07_resultB*\n(_current_state_progress_hundreth_percentB\x0e\n\x0c_transfer_idB\x12\n\x10_bundle_filenameB\x07\n\x05_noteB\x12\n\x10_extended_status\"@\n\x0cStartRequest\x12\x1c\n\x0f\x62undle_filename\x18\x01 \x01(\tH\x00\x88\x01\x01\x42\x12\n\x10_bundle_filename2\x95\x04\n\rBundledUpdate\x12H\n\tGetStatus\x12\x12.pw.protobuf.Empty\x1a\'.pw.software_update.BundledUpdateStatus\x12R\n\x05Start\x12 .pw.software_update.StartRequest\x1a\'.pw.software_update.BundledUpdateStatus\x12M\n\x0eSetTransferred\x12\x12.pw.protobuf.Empty\x1a\'.pw.software_update.BundledUpdateStatus\x12\x45\n\x06Verify\x12\x12.pw.protobuf.Empty\x1a\'.pw.software_update.BundledUpdateStatus\x12\x44\n\x05\x41pply\x12\x12.pw.protobuf.Empty\x1a\'.pw.software_update.BundledUpdateStatus\x12\x44\n\x05\x41\x62ort\x12\x12.pw.protobuf.Empty\x1a\'.pw.software_update.BundledUpdateStatus\x12\x44\n\x05Reset\x12\x12.pw.protobuf.Empty\x1a\'.pw.software_update.BundledUpdateStatusb\x06proto3'
  ,
  dependencies=[pw__protobuf__protos_dot_common__pb2.DESCRIPTOR,pw__tokenizer_dot_proto_dot_options__pb2.DESCRIPTOR,google_dot_protobuf_dot_any__pb2.DESCRIPTOR,])



_BUNDLEDUPDATESTATE_ENUM = _descriptor.EnumDescriptor(
  name='Enum',
  full_name='pw.software_update.BundledUpdateState.Enum',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INACTIVE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TRANSFERRING', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TRANSFERRED', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='VERIFYING', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='VERIFIED', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='APPLYING', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ABORTING', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FINISHED', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=181,
  serialized_end=320,
)
_sym_db.RegisterEnumDescriptor(_BUNDLEDUPDATESTATE_ENUM)

_BUNDLEDUPDATERESULT_ENUM = _descriptor.EnumDescriptor(
  name='Enum',
  full_name='pw.software_update.BundledUpdateResult.Enum',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SUCCESS', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN_ERROR', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ABORTED', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TRANSFER_FAILED', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='VERIFY_FAILED', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='APPLY_FAILED', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=346,
  serialized_end=468,
)
_sym_db.RegisterEnumDescriptor(_BUNDLEDUPDATERESULT_ENUM)


_BUNDLEDUPDATESTATE = _descriptor.Descriptor(
  name='BundledUpdateState',
  full_name='pw.software_update.BundledUpdateState',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _BUNDLEDUPDATESTATE_ENUM,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=158,
  serialized_end=320,
)


_BUNDLEDUPDATERESULT = _descriptor.Descriptor(
  name='BundledUpdateResult',
  full_name='pw.software_update.BundledUpdateResult',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _BUNDLEDUPDATERESULT_ENUM,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=323,
  serialized_end=468,
)


_BUNDLEDUPDATESTATUS = _descriptor.Descriptor(
  name='BundledUpdateStatus',
  full_name='pw.software_update.BundledUpdateStatus',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='state', full_name='pw.software_update.BundledUpdateStatus.state', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='result', full_name='pw.software_update.BundledUpdateStatus.result', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='current_state_progress_hundreth_percent', full_name='pw.software_update.BundledUpdateStatus.current_state_progress_hundreth_percent', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='transfer_id', full_name='pw.software_update.BundledUpdateStatus.transfer_id', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bundle_filename', full_name='pw.software_update.BundledUpdateStatus.bundle_filename', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='note', full_name='pw.software_update.BundledUpdateStatus.note', index=5,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\200\257&\000', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='extended_status', full_name='pw.software_update.BundledUpdateStatus.extended_status', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_result', full_name='pw.software_update.BundledUpdateStatus._result',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_current_state_progress_hundreth_percent', full_name='pw.software_update.BundledUpdateStatus._current_state_progress_hundreth_percent',
      index=1, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_transfer_id', full_name='pw.software_update.BundledUpdateStatus._transfer_id',
      index=2, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_bundle_filename', full_name='pw.software_update.BundledUpdateStatus._bundle_filename',
      index=3, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_note', full_name='pw.software_update.BundledUpdateStatus._note',
      index=4, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
    _descriptor.OneofDescriptor(
      name='_extended_status', full_name='pw.software_update.BundledUpdateStatus._extended_status',
      index=5, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=471,
  serialized_end=926,
)


_STARTREQUEST = _descriptor.Descriptor(
  name='StartRequest',
  full_name='pw.software_update.StartRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='bundle_filename', full_name='pw.software_update.StartRequest.bundle_filename', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='_bundle_filename', full_name='pw.software_update.StartRequest._bundle_filename',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=928,
  serialized_end=992,
)

_BUNDLEDUPDATESTATE_ENUM.containing_type = _BUNDLEDUPDATESTATE
_BUNDLEDUPDATERESULT_ENUM.containing_type = _BUNDLEDUPDATERESULT
_BUNDLEDUPDATESTATUS.fields_by_name['state'].enum_type = _BUNDLEDUPDATESTATE_ENUM
_BUNDLEDUPDATESTATUS.fields_by_name['result'].enum_type = _BUNDLEDUPDATERESULT_ENUM
_BUNDLEDUPDATESTATUS.fields_by_name['extended_status'].message_type = google_dot_protobuf_dot_any__pb2._ANY
_BUNDLEDUPDATESTATUS.oneofs_by_name['_result'].fields.append(
  _BUNDLEDUPDATESTATUS.fields_by_name['result'])
_BUNDLEDUPDATESTATUS.fields_by_name['result'].containing_oneof = _BUNDLEDUPDATESTATUS.oneofs_by_name['_result']
_BUNDLEDUPDATESTATUS.oneofs_by_name['_current_state_progress_hundreth_percent'].fields.append(
  _BUNDLEDUPDATESTATUS.fields_by_name['current_state_progress_hundreth_percent'])
_BUNDLEDUPDATESTATUS.fields_by_name['current_state_progress_hundreth_percent'].containing_oneof = _BUNDLEDUPDATESTATUS.oneofs_by_name['_current_state_progress_hundreth_percent']
_BUNDLEDUPDATESTATUS.oneofs_by_name['_transfer_id'].fields.append(
  _BUNDLEDUPDATESTATUS.fields_by_name['transfer_id'])
_BUNDLEDUPDATESTATUS.fields_by_name['transfer_id'].containing_oneof = _BUNDLEDUPDATESTATUS.oneofs_by_name['_transfer_id']
_BUNDLEDUPDATESTATUS.oneofs_by_name['_bundle_filename'].fields.append(
  _BUNDLEDUPDATESTATUS.fields_by_name['bundle_filename'])
_BUNDLEDUPDATESTATUS.fields_by_name['bundle_filename'].containing_oneof = _BUNDLEDUPDATESTATUS.oneofs_by_name['_bundle_filename']
_BUNDLEDUPDATESTATUS.oneofs_by_name['_note'].fields.append(
  _BUNDLEDUPDATESTATUS.fields_by_name['note'])
_BUNDLEDUPDATESTATUS.fields_by_name['note'].containing_oneof = _BUNDLEDUPDATESTATUS.oneofs_by_name['_note']
_BUNDLEDUPDATESTATUS.oneofs_by_name['_extended_status'].fields.append(
  _BUNDLEDUPDATESTATUS.fields_by_name['extended_status'])
_BUNDLEDUPDATESTATUS.fields_by_name['extended_status'].containing_oneof = _BUNDLEDUPDATESTATUS.oneofs_by_name['_extended_status']
_STARTREQUEST.oneofs_by_name['_bundle_filename'].fields.append(
  _STARTREQUEST.fields_by_name['bundle_filename'])
_STARTREQUEST.fields_by_name['bundle_filename'].containing_oneof = _STARTREQUEST.oneofs_by_name['_bundle_filename']
DESCRIPTOR.message_types_by_name['BundledUpdateState'] = _BUNDLEDUPDATESTATE
DESCRIPTOR.message_types_by_name['BundledUpdateResult'] = _BUNDLEDUPDATERESULT
DESCRIPTOR.message_types_by_name['BundledUpdateStatus'] = _BUNDLEDUPDATESTATUS
DESCRIPTOR.message_types_by_name['StartRequest'] = _STARTREQUEST
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

BundledUpdateState = _reflection.GeneratedProtocolMessageType('BundledUpdateState', (_message.Message,), {
  'DESCRIPTOR' : _BUNDLEDUPDATESTATE,
  '__module__' : 'pw_software_update.bundled_update_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.BundledUpdateState)
  })
_sym_db.RegisterMessage(BundledUpdateState)

BundledUpdateResult = _reflection.GeneratedProtocolMessageType('BundledUpdateResult', (_message.Message,), {
  'DESCRIPTOR' : _BUNDLEDUPDATERESULT,
  '__module__' : 'pw_software_update.bundled_update_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.BundledUpdateResult)
  })
_sym_db.RegisterMessage(BundledUpdateResult)

BundledUpdateStatus = _reflection.GeneratedProtocolMessageType('BundledUpdateStatus', (_message.Message,), {
  'DESCRIPTOR' : _BUNDLEDUPDATESTATUS,
  '__module__' : 'pw_software_update.bundled_update_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.BundledUpdateStatus)
  })
_sym_db.RegisterMessage(BundledUpdateStatus)

StartRequest = _reflection.GeneratedProtocolMessageType('StartRequest', (_message.Message,), {
  'DESCRIPTOR' : _STARTREQUEST,
  '__module__' : 'pw_software_update.bundled_update_pb2'
  # @@protoc_insertion_point(class_scope:pw.software_update.StartRequest)
  })
_sym_db.RegisterMessage(StartRequest)


_BUNDLEDUPDATESTATUS.fields_by_name['note']._options = None

_BUNDLEDUPDATE = _descriptor.ServiceDescriptor(
  name='BundledUpdate',
  full_name='pw.software_update.BundledUpdate',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=995,
  serialized_end=1528,
  methods=[
  _descriptor.MethodDescriptor(
    name='GetStatus',
    full_name='pw.software_update.BundledUpdate.GetStatus',
    index=0,
    containing_service=None,
    input_type=pw__protobuf__protos_dot_common__pb2._EMPTY,
    output_type=_BUNDLEDUPDATESTATUS,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Start',
    full_name='pw.software_update.BundledUpdate.Start',
    index=1,
    containing_service=None,
    input_type=_STARTREQUEST,
    output_type=_BUNDLEDUPDATESTATUS,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='SetTransferred',
    full_name='pw.software_update.BundledUpdate.SetTransferred',
    index=2,
    containing_service=None,
    input_type=pw__protobuf__protos_dot_common__pb2._EMPTY,
    output_type=_BUNDLEDUPDATESTATUS,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Verify',
    full_name='pw.software_update.BundledUpdate.Verify',
    index=3,
    containing_service=None,
    input_type=pw__protobuf__protos_dot_common__pb2._EMPTY,
    output_type=_BUNDLEDUPDATESTATUS,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Apply',
    full_name='pw.software_update.BundledUpdate.Apply',
    index=4,
    containing_service=None,
    input_type=pw__protobuf__protos_dot_common__pb2._EMPTY,
    output_type=_BUNDLEDUPDATESTATUS,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Abort',
    full_name='pw.software_update.BundledUpdate.Abort',
    index=5,
    containing_service=None,
    input_type=pw__protobuf__protos_dot_common__pb2._EMPTY,
    output_type=_BUNDLEDUPDATESTATUS,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Reset',
    full_name='pw.software_update.BundledUpdate.Reset',
    index=6,
    containing_service=None,
    input_type=pw__protobuf__protos_dot_common__pb2._EMPTY,
    output_type=_BUNDLEDUPDATESTATUS,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_BUNDLEDUPDATE)

DESCRIPTOR.services_by_name['BundledUpdate'] = _BUNDLEDUPDATE

# @@protoc_insertion_point(module_scope)
